import operator
from datetime import timedelta

from pyasn1_alt_modules import rfc5280, rfc6962, rfc5480

import pkilint.common
from pkilint import validation, document, oid, common
from pkilint.cabf import cabf_name
from pkilint.cabf.asn1 import ev_guidelines
from pkilint.cabf.serverauth import serverauth_constants
from pkilint.common import organization_id, common_name
from pkilint.common.organization_id import ParsedOrganizationIdentifier
from pkilint.itu import x520_name, bitstring
from pkilint.pkix import Rfc2119Word, general_name, time
from pkilint.pkix.certificate.certificate_extension import KeyUsageBitName


def _parse_organization_identifier_extension(
    ext_node: document.PDUNode,
) -> organization_id.ParsedOrganizationIdentifier:
    state_province_node = ext_node.children.get("registrationStateOrProvince")

    state_province = str(state_province_node.pdu) if state_province_node else None

    reference_node = ext_node.children.get("registrationReference")
    reference = str(reference_node.pdu) if reference_node else None

    return organization_id.ParsedOrganizationIdentifier(
        raw=None,
        scheme=str(ext_node.children["registrationSchemeIdentifier"].pdu),
        is_national_scheme=False,
        country=str(ext_node.children["registrationCountry"].pdu),
        state_province=state_province,
        reference=reference,
    )


class CABFOrganizationIdentifierExtensionValidator(
    cabf_name.CabfOrganizationIdentifierValidatorBase
):
    """Validates that the content of the CA/B Forum organizationIdentifier extension conforms with EVG 9.8.2."""

    def __init__(self):
        super().__init__(
            invalid_format_validation=None,
            enforce_strict_state_province_format=True,
            pdu_class=ev_guidelines.CABFOrganizationIdentifier,
        )

    @classmethod
    def parse_organization_id_node(
        cls, node: document.PDUNode
    ) -> ParsedOrganizationIdentifier:
        return _parse_organization_identifier_extension(node)


class EvSubscriberAttributeAllowanceValidator(
    pkilint.common.AttributeIdentifierAllowanceValidator
):
    """Validates that the content of the subject conforms to EVG 9.2."""

    _CODE_CLASSIFIER = "cabf.ev_guidelines"

    _ATTRIBUTE_ALLOWANCES = {
        rfc5280.id_at_countryName: Rfc2119Word.MUST,
        rfc5280.id_at_stateOrProvinceName: Rfc2119Word.MAY,
        rfc5280.id_at_localityName: Rfc2119Word.MAY,
        rfc5280.id_at_serialNumber: Rfc2119Word.MUST,
        x520_name.id_at_businessCategory: Rfc2119Word.MUST,
        rfc5280.id_at_organizationName: Rfc2119Word.MUST,
        x520_name.id_at_organizationIdentifier: Rfc2119Word.MAY,
        x520_name.id_at_postalCode: Rfc2119Word.MAY,
        x520_name.id_at_streetAddress: Rfc2119Word.MAY,
        ev_guidelines.id_evat_jurisdiction_countryName: Rfc2119Word.MUST,
        ev_guidelines.id_evat_jurisdiction_stateOrProvinceName: Rfc2119Word.MAY,
        ev_guidelines.id_evat_jurisdiction_localityName: Rfc2119Word.MAY,
        rfc5280.id_at_commonName: Rfc2119Word.SHOULD_NOT,
    }

    def __init__(self):
        super().__init__(
            self._ATTRIBUTE_ALLOWANCES, self._CODE_CLASSIFIER, Rfc2119Word.MUST_NOT
        )


class DvSubcriberAttributeAllowanceValidator(
    pkilint.common.AttributeIdentifierAllowanceValidator
):
    """Validates that the content of the subject field conforms with BR 7.1.7.2."""

    _CODE_CLASSIFIER = "cabf.serverauth.dv"

    _ATTRIBUTE_ALLOWANCES = {
        rfc5280.id_at_countryName: Rfc2119Word.MAY,
        rfc5280.id_at_commonName: Rfc2119Word.SHOULD_NOT,
    }

    def __init__(self):
        super().__init__(
            self._ATTRIBUTE_ALLOWANCES, self._CODE_CLASSIFIER, Rfc2119Word.MUST_NOT
        )


class IvSubscriberAttributeAllowanceValidator(
    pkilint.common.AttributeIdentifierAllowanceValidator
):
    """Validates that the content of the subject field conforms with BR 7.1.7.3."""

    _CODE_CLASSIFIER = "cabf.serverauth.iv"

    _ATTRIBUTE_ALLOWANCES = {
        rfc5280.id_at_countryName: Rfc2119Word.MUST,
        rfc5280.id_at_stateOrProvinceName: Rfc2119Word.MAY,
        rfc5280.id_at_localityName: Rfc2119Word.MAY,
        x520_name.id_at_postalCode: Rfc2119Word.SHOULD_NOT,
        x520_name.id_at_streetAddress: Rfc2119Word.SHOULD_NOT,
        rfc5280.id_at_organizationName: Rfc2119Word.SHOULD_NOT,
        rfc5280.id_at_surname: Rfc2119Word.MUST,
        rfc5280.id_at_givenName: Rfc2119Word.MUST,
        rfc5280.id_at_organizationalUnitName: Rfc2119Word.MUST_NOT,
        rfc5280.id_at_commonName: Rfc2119Word.SHOULD_NOT,
    }

    def __init__(self):
        super().__init__(
            self._ATTRIBUTE_ALLOWANCES, self._CODE_CLASSIFIER, Rfc2119Word.SHOULD_NOT
        )


class OvSubscriberAttributeAllowanceValidator(
    pkilint.common.AttributeIdentifierAllowanceValidator
):
    """Validates that the content of the subject field conforms with BR 7.1.7.4."""

    _CODE_CLASSIFIER = "cabf.serverauth.ov"

    _ATTRIBUTE_ALLOWANCES = {
        rfc5280.id_domainComponent: Rfc2119Word.MAY,
        rfc5280.id_at_countryName: Rfc2119Word.MUST,
        rfc5280.id_at_stateOrProvinceName: Rfc2119Word.MAY,
        rfc5280.id_at_localityName: Rfc2119Word.MAY,
        x520_name.id_at_postalCode: Rfc2119Word.SHOULD_NOT,
        x520_name.id_at_streetAddress: Rfc2119Word.SHOULD_NOT,
        rfc5280.id_at_organizationName: Rfc2119Word.MUST,
        rfc5280.id_at_surname: Rfc2119Word.MUST_NOT,
        rfc5280.id_at_givenName: Rfc2119Word.MUST_NOT,
        rfc5280.id_at_organizationalUnitName: Rfc2119Word.MUST_NOT,
        rfc5280.id_at_commonName: Rfc2119Word.SHOULD_NOT,
    }

    def __init__(self):
        super().__init__(
            self._ATTRIBUTE_ALLOWANCES, self._CODE_CLASSIFIER, Rfc2119Word.SHOULD_NOT
        )


# EVG 9.2.6, BR 7.1.2.7.3, BR 7.1.2.7.4
class IdentityCertificateStateProvinceAndLocalityPresenceValidator(
    validation.Validator
):
    """Validates that the stateOrProvinceName and/or localityName subject attributes are present, as per
    EVG 9.2.6, BR 7.1.2.7.3, and BR 7.1.2.7.4."""

    VALIDATION_STATEPROVINCE_AND_LOCALITY_MISSING = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "cabf.serverauth.subscriber_stateprovince_and_locality_missing",
    )

    def __init__(self):
        super().__init__(
            validations=self.VALIDATION_STATEPROVINCE_AND_LOCALITY_MISSING,
            path="certificate.tbsCertificate.subject",
        )

    def validate(self, node):
        stp_attr = node.document.get_subject_attributes_by_type(
            rfc5280.id_at_stateOrProvinceName
        )
        loc_attr = node.document.get_subject_attributes_by_type(
            rfc5280.id_at_localityName
        )

        if not any(stp_attr) and not any(loc_attr):
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_STATEPROVINCE_AND_LOCALITY_MISSING
            )


class EvSubscriberJurisdictionPresenceValidator(validation.Validator):
    """Validates that jurisdictionStateOrProvinceName is present when jurisdictionLocalityName is present,
    as per EVG 9.2.4."""

    VALIDATION_JURIS_STP_ABSENT_LOCALITY_PRESENT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "cabf.ev_guidelines.jurisdiction_locality_present_stateprovince_missing",
    )

    def __init__(self):
        super().__init__(
            validations=self.VALIDATION_JURIS_STP_ABSENT_LOCALITY_PRESENT,
            path="certificate.tbsCertificate.subject",
        )

    def validate(self, node):
        juris_stp_attr = node.document.get_subject_attributes_by_type(
            ev_guidelines.id_evat_jurisdiction_stateOrProvinceName
        )
        juris_loc_attr = node.document.get_subject_attributes_by_type(
            ev_guidelines.id_evat_jurisdiction_localityName
        )

        if any(juris_loc_attr) and not any(juris_stp_attr):
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_JURIS_STP_ABSENT_LOCALITY_PRESENT
            )


class SubscriberExtensionAllowanceValidator(
    pkilint.common.ExtensionIdentifierAllowanceValidator
):
    """Validates that the included extensions conform with BR 7.1.2.7.6."""

    _CODE_CLASSIFIER = "cabf.serverauth.subscriber"

    _EXTENSION_ALLOWANCES = {
        rfc5280.id_pe_authorityInfoAccess: Rfc2119Word.MUST,
        rfc5280.id_ce_authorityKeyIdentifier: Rfc2119Word.MUST,
        rfc5280.id_ce_certificatePolicies: Rfc2119Word.MUST,
        rfc5280.id_ce_extKeyUsage: Rfc2119Word.MUST,
        rfc5280.id_ce_subjectAltName: Rfc2119Word.MUST,
        rfc5280.id_ce_nameConstraints: Rfc2119Word.MUST_NOT,
        rfc5280.id_ce_keyUsage: Rfc2119Word.SHOULD,
        rfc5280.id_ce_basicConstraints: Rfc2119Word.MAY,
        rfc5280.id_ce_cRLDistributionPoints: Rfc2119Word.MAY,
        rfc5280.id_ce_subjectKeyIdentifier: Rfc2119Word.SHOULD_NOT,
    }

    def __init__(self, certificate_type):
        allowances = self._EXTENSION_ALLOWANCES.copy()

        if certificate_type in serverauth_constants.SUBSCRIBER_PRECERT_TYPES:
            allowances.update(
                {
                    rfc6962.id_ce_criticalPoison: Rfc2119Word.MUST,
                    rfc6962.id_ce_embeddedSCT: Rfc2119Word.MUST_NOT,
                }
            )
        elif (
            certificate_type in serverauth_constants.SUBSCRIBER_FINAL_CERTIFICATE_TYPES
        ):
            allowances.update(
                {
                    rfc6962.id_ce_criticalPoison: Rfc2119Word.MUST_NOT,
                    rfc6962.id_ce_embeddedSCT: Rfc2119Word.MAY,
                }
            )
        else:
            raise ValueError(f"Unsupported certificate type: {certificate_type}")

        super().__init__(allowances, self._CODE_CLASSIFIER, Rfc2119Word.SHOULD_NOT)


class SubscriberEkuAllowanceValidator(
    pkilint.common.ExtendedKeyUsageAllowanceValidator
):
    """Validates that the content of the extended key usage extension conforms with BR 7.1.2.7.10."""

    _CODE_CLASSIFIER = "cabf.serverauth.subscriber"

    _EKU_ALLOWANCES = {
        rfc5280.id_kp_serverAuth: Rfc2119Word.MUST,
        rfc5280.id_kp_clientAuth: Rfc2119Word.MAY,
        rfc5280.id_kp_codeSigning: Rfc2119Word.MUST_NOT,
        rfc5280.id_kp_emailProtection: Rfc2119Word.MUST_NOT,
        rfc5280.id_kp_timeStamping: Rfc2119Word.MUST_NOT,
        rfc5280.id_kp_OCSPSigning: Rfc2119Word.MUST_NOT,
        rfc5280.anyExtendedKeyUsage: Rfc2119Word.MUST_NOT,
        rfc6962.id_kp_precertificateSigning: Rfc2119Word.MUST_NOT,
    }

    def __init__(self):
        super().__init__(
            self._EKU_ALLOWANCES, self._CODE_CLASSIFIER, Rfc2119Word.SHOULD_NOT
        )


class SubscriberKeyUsageValidator(validation.Validator):
    """Validates that the content of the key usage extension conforms with BR 7.1.2.7.11."""

    VALIDATION_REQUIRED_KU_MISSING = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "cabf.serverauth.subscriber_required_ku_missing",
    )

    VALIDATION_RECOMMENDED_KU_MISSING = validation.ValidationFinding(
        validation.ValidationFindingSeverity.WARNING,
        "cabf.serverauth.subscriber_recommended_ku_missing",
    )

    VALIDATION_PROHIBITED_KU_PRESENT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "cabf.serverauth.subscriber_prohibited_ku_present",
    )

    VALIDATION_DISCOURAGED_KU_PRESENT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.WARNING,
        "cabf.serverauth.subscriber_discouraged_ku_present",
    )

    VALIDATION_RSA_DIGSIG_AND_KEYENCIPHERMENT_PRESENT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.WARNING,
        "cabf.serverauth.subscriber_rsa_digitalsignature_and_keyencipherment_present",
    )

    _SPKI_OID_TO_KU_ALLOWANCES_MAPPING = {
        rfc5480.rsaEncryption: {
            KeyUsageBitName.DIGITAL_SIGNATURE: Rfc2119Word.SHOULD,
            KeyUsageBitName.KEY_ENCIPHERMENT: Rfc2119Word.MAY,
            KeyUsageBitName.DATA_ENCIPHERMENT: Rfc2119Word.SHOULD_NOT,
        },
        rfc5480.id_ecPublicKey: {
            KeyUsageBitName.DIGITAL_SIGNATURE: Rfc2119Word.MUST,
            KeyUsageBitName.KEY_AGREEMENT: Rfc2119Word.SHOULD_NOT,
        },
    }

    def __init__(self):
        super().__init__(
            validations=[
                self.VALIDATION_REQUIRED_KU_MISSING,
                self.VALIDATION_PROHIBITED_KU_PRESENT,
                self.VALIDATION_RECOMMENDED_KU_MISSING,
                self.VALIDATION_DISCOURAGED_KU_PRESENT,
                self.VALIDATION_RSA_DIGSIG_AND_KEYENCIPHERMENT_PRESENT,
            ],
            pdu_class=rfc5280.KeyUsage,
        )

    def validate(self, node):
        spki_alg_oid = node.navigate(
            ":certificate.tbsCertificate.subjectPublicKeyInfo.algorithm.algorithm"
        ).pdu

        allowances = self._SPKI_OID_TO_KU_ALLOWANCES_MAPPING.get(spki_alg_oid)

        if allowances is None:
            # unsupported SPKI, just return and let another validator throw the error
            return

        warning_findings = []

        for k in rfc5280.KeyUsage.namedValues:
            k = str(k)

            requirement_word = allowances.get(k)

            if bitstring.has_named_bit(node, k):
                if requirement_word is None:
                    raise validation.ValidationFindingEncountered(
                        self.VALIDATION_PROHIBITED_KU_PRESENT,
                        f"Prohibited KU present: {k}",
                    )
                elif requirement_word == Rfc2119Word.SHOULD_NOT:
                    warning_findings.append(
                        validation.ValidationFindingDescription(
                            self.VALIDATION_DISCOURAGED_KU_PRESENT,
                            f"Discouraged KU present: {k}",
                        )
                    )
            else:
                if requirement_word == Rfc2119Word.MUST:
                    raise validation.ValidationFindingEncountered(
                        self.VALIDATION_REQUIRED_KU_MISSING, f"Required KU missing: {k}"
                    )
                elif requirement_word == Rfc2119Word.SHOULD:
                    warning_findings.append(
                        validation.ValidationFindingDescription(
                            self.VALIDATION_RECOMMENDED_KU_MISSING,
                            f"Recommended KU missing: {k}",
                        )
                    )

        if spki_alg_oid == rfc5480.rsaEncryption and (
            bitstring.has_named_bit(node, KeyUsageBitName.DIGITAL_SIGNATURE)
            and bitstring.has_named_bit(node, KeyUsageBitName.KEY_ENCIPHERMENT)
        ):
            warning_findings.append(
                validation.ValidationFindingDescription(
                    self.VALIDATION_RSA_DIGSIG_AND_KEYENCIPHERMENT_PRESENT, None
                )
            )

        return validation.ValidationResult(self, node, warning_findings)


class SubscriberSanGeneralNameTypeValidator(validation.Validator):
    """Validates that the types of GeneralNames included in the SAN extension conform to BR 7.1.2.7.12."""

    VALIDATION_PROHIBITED_SAN_TYPE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "cabf.serverauth.prohibited_san_type",
    )

    def __init__(self):
        super().__init__(
            validations=self.VALIDATION_PROHIBITED_SAN_TYPE,
            pdu_class=rfc5280.GeneralName,
            predicate=lambda n: n.parent is not None
            and isinstance(n.parent.pdu, rfc5280.SubjectAltName),
        )

    def validate(self, node):
        gn_type, _ = node.child

        if gn_type not in {"dNSName", "iPAddress"}:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_PROHIBITED_SAN_TYPE,
                f"Prohibited SAN GeneralName type: {gn_type}",
            )


class EvSanGeneralNameTypeValidator(validation.Validator):
    """Validates that the types of GeneralNames included in the SAN extension conform to EVG 9.8.1."""

    VALIDATION_PROHIBITED_SAN_TYPE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "cabf.ev_guidelines.prohibited_san_type",
    )

    def __init__(self):
        super().__init__(
            validations=self.VALIDATION_PROHIBITED_SAN_TYPE,
            pdu_class=rfc5280.GeneralName,
            predicate=lambda n: n.parent is not None
            and isinstance(n.parent.pdu, rfc5280.SubjectAltName),
        )

    def validate(self, node):
        gn_type, _ = node.child

        if gn_type != "dNSName":
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_PROHIBITED_SAN_TYPE,
                f"Prohibited SAN GeneralName type: {gn_type}",
            )


class SubscriberValidityPeriodValidator(time.ValidityPeriodThresholdsValidator):
    """Validates that the validity period conforms to BR 7.1.2.7."""

    VALIDATION_VALIDITY_PERIOD_EXCEEDS_398_DAYS = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "cabf.certificate_validity_period_exceeds_398_days",
    )

    VALIDATION_VALIDITY_PERIOD_EXCEEDS_397_DAYS = validation.ValidationFinding(
        validation.ValidationFindingSeverity.WARNING,
        "cabf.certificate_validity_period_exceeds_397_days",
    )

    _THRESHOLDS = [
        (operator.le, timedelta(days=398), VALIDATION_VALIDITY_PERIOD_EXCEEDS_398_DAYS),
        (operator.le, timedelta(days=397), VALIDATION_VALIDITY_PERIOD_EXCEEDS_397_DAYS),
    ]

    def __init__(self):
        super().__init__(
            end_validity_node_retriever=lambda n: n.navigate("^.notAfter"),
            inclusive_second=True,
            validity_period_thresholds=self._THRESHOLDS,
            path="certificate.tbsCertificate.validity.notBefore",
        )


class SubscriberCommonNameValidator(common_name.CommonNameValidator):
    """Validates that the content of the commonName attribute conforms to BR 7.1.4.3."""

    VALIDATION_COMMON_NAME_UNKNOWN_SOURCE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "cabf.serverauth.subscriber_common_name_unknown_source",
    )

    def __init__(self):
        super().__init__(
            {
                general_name.GeneralNameTypeName.DNS_NAME,
                general_name.GeneralNameTypeName.IP_ADDRESS,
            },
            self.VALIDATION_COMMON_NAME_UNKNOWN_SOURCE,
        )


class SubscriberExtensionCriticalityValidator(
    pkilint.common.ExtensionCriticalityValidator
):
    """Validates that the criticality of extensions conforms to BR 7.1.2.7.6."""

    _CODE_CLASSIFIER = "cabf.serverauth.subscriber"

    _CRITICALITY_MAPPING = {
        rfc5280.id_pe_authorityInfoAccess: False,
        rfc5280.id_ce_authorityKeyIdentifier: False,
        rfc5280.id_ce_certificatePolicies: False,
        rfc5280.id_ce_extKeyUsage: False,
        rfc5280.id_ce_keyUsage: True,
        rfc5280.id_ce_basicConstraints: True,
        rfc5280.id_ce_cRLDistributionPoints: False,
        rfc6962.id_ce_embeddedSCT: False,
        rfc5280.id_ce_subjectKeyIdentifier: False,
    }

    def __init__(self):
        super().__init__(
            self._CRITICALITY_MAPPING,
            self._CODE_CLASSIFIER,
            Rfc2119Word.MUST,
            Rfc2119Word.MUST,
        )


class SubscriberBasicConstraintsValidator(validation.Validator):
    VALIDATION_CA_BIT_SET = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "cabf.serverauth.subscriber_basic_constraints_ca_bit_set",
    )

    def __init__(self):
        super().__init__(
            validations=self.VALIDATION_CA_BIT_SET, pdu_class=rfc5280.BasicConstraints
        )

    def validate(self, node):
        if bool(node.children["cA"].pdu):
            raise validation.ValidationFindingEncountered(self.VALIDATION_CA_BIT_SET)


class SubscriberPoliciesValidator(validation.Validator):
    """Validates that the certificate policy OID(s) conform to BR 7.1.2.7.9."""

    VALIDATION_ANYPOLICY_PRESENT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "cabf.serverauth.subscriber_anypolicy_oid_present",
    )

    VALIDATION_MULTIPLE_RESERVED_OIDS = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "cabf.serverauth.subscriber_multiple_reserved_policy_oids",
    )

    VALIDATION_NO_RESERVED_OID = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "cabf.serverauth.subscriber_missing_reserved_policy_oid",
    )

    VALIDATION_FIRST_OID_NOT_RESERVED = validation.ValidationFinding(
        validation.ValidationFindingSeverity.WARNING,
        "cabf.serverauth.subscriber_first_policy_oid_not_reserved",
    )

    def __init__(self, certificate_type: serverauth_constants.CertificateType):
        if certificate_type not in serverauth_constants.SUBSCRIBER_CERTIFICATE_TYPES:
            raise ValueError(f"Unsupported certificate type: {certificate_type}")

        self._certificate_type = certificate_type

        if certificate_type in serverauth_constants.EV_CERTIFICATE_TYPES:
            self._expected_reserved_oid = serverauth_constants.ID_POLICY_EV
        elif certificate_type in serverauth_constants.OV_CERTIFICATE_TYPES:
            self._expected_reserved_oid = serverauth_constants.ID_POLICY_OV
        elif certificate_type in serverauth_constants.IV_CERTIFICATE_TYPES:
            self._expected_reserved_oid = serverauth_constants.ID_POLICY_IV
        elif certificate_type in serverauth_constants.DV_CERTIFICATE_TYPES:
            self._expected_reserved_oid = serverauth_constants.ID_POLICY_DV
        else:
            raise ValueError(f"Unsupported certificate type: {certificate_type}")

        super().__init__(
            validations=[
                self.VALIDATION_ANYPOLICY_PRESENT,
                self.VALIDATION_MULTIPLE_RESERVED_OIDS,
                self.VALIDATION_NO_RESERVED_OID,
                self.VALIDATION_FIRST_OID_NOT_RESERVED,
            ],
            pdu_class=rfc5280.CertificatePolicies,
        )

    def validate(self, node):
        policy_oids = [
            pi.children["policyIdentifier"].pdu for pi in node.children.values()
        ]

        if rfc5280.anyPolicy in policy_oids:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_ANYPOLICY_PRESENT
            )

        if self._expected_reserved_oid not in policy_oids:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_NO_RESERVED_OID,
                f'Required policy OID "{str(self._expected_reserved_oid)}" missing',
            )

        reserved_oids = (
            set(policy_oids) & serverauth_constants.SERVERAUTH_RESERVED_POLICY_OIDS
        )

        if len(reserved_oids) > 1:
            oids_str = oid.format_oids(reserved_oids)

            raise validation.ValidationFindingEncountered(
                self.VALIDATION_MULTIPLE_RESERVED_OIDS,
                f"Multiple reserved policy OIDs present: {oids_str}",
            )

        if policy_oids[0] not in serverauth_constants.SERVERAUTH_RESERVED_POLICY_OIDS:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_FIRST_OID_NOT_RESERVED
            )


class EvWildcardAllowanceValidator(validation.Validator):
    """Validates that wildcard dNSNames conform to EVG 9.8.1."""

    VALIDATION_EV_WILDCARD_SAN_PRESENT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "cabf.ev_guidelines.ev_wildcard_san_present",
    )

    def __init__(self):
        super().__init__(
            validations=self.VALIDATION_EV_WILDCARD_SAN_PRESENT,
            predicate=general_name.create_generalname_type_predicate("dNSName"),
        )

    def validate(self, node):
        value = str(node.pdu)

        if value.startswith("*") and not value.lower().endswith(".onion"):
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_EV_WILDCARD_SAN_PRESENT,
                f'Wildcard SAN present: "{value}"',
            )


class SubscriberAuthorityInformationAccessAccessMethodPresenceValidator(
    common.AuthorityInformationAccessAccessMethodPresenceValidator
):
    """Validates that AIA access methods conform to BR 7.1.2.10.3."""

    _CODE_CLASSIFIER = "cabf.serverauth.subscriber"

    _ACCESS_METHOD_ALLOWANCES = {
        rfc5280.id_ad_ocsp: Rfc2119Word.MUST,
        rfc5280.id_ad_caIssuers: Rfc2119Word.SHOULD,
    }

    def __init__(self):
        super().__init__(
            self._ACCESS_METHOD_ALLOWANCES, self._CODE_CLASSIFIER, Rfc2119Word.MUST_NOT
        )


class OrganizationIdentifierConsistentSubjectAndExtensionValidator(
    validation.Validator
):
    """Validates that the content of the organizationIdentifier subject attributes and the organizationIdentifier
    extension are consistent, as per EVG 9.2.8 and 9.2.9."""

    VALIDATION_CABF_ORG_ID_NO_EXT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "cabf.serverauth.organization_identifier_extension_absent",
    )

    VALIDATION_CABF_ORG_ID_MISMATCHED_VALUE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "cabf.serverauth.organization_identifier_mismatched_value",
    )

    def __init__(self):
        super().__init__(
            validations=[
                self.VALIDATION_CABF_ORG_ID_NO_EXT,
                self.VALIDATION_CABF_ORG_ID_MISMATCHED_VALUE,
            ],
            pdu_class=x520_name.X520OrganizationIdentifier,
            predicate=lambda n: any(n.children),
        )

    def validate(self, node):
        ext_and_idx = node.document.get_extension_by_oid(
            ev_guidelines.id_CABFOrganizationIdentifier
        )

        if ext_and_idx is None:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_CABF_ORG_ID_NO_EXT
            )

        ext_node, _ = ext_and_idx
        try:
            ext_node = ext_node.navigate("extnValue.cABFOrganizationIdentifier")
        except document.PDUNavigationFailedError:
            return

        attr_value = str(node.child[1].pdu)

        try:
            org_id_attr_parsed = organization_id.parse_organization_identifier(
                attr_value
            )
        except ValueError:
            # let the format validator report this error
            return

        org_id_ext_parsed = _parse_organization_identifier_extension(ext_node)

        try:
            organization_id.assert_parsed_organization_identifier_equal(
                org_id_attr_parsed, "attribute", org_id_ext_parsed, "extension"
            )
        except ValueError as e:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_CABF_ORG_ID_MISMATCHED_VALUE, str(e)
            )



import enum
from enum import auto

from pkilint.cabf.serverauth import serverauth_constants

EN_319_412_VERSION = "2023-09"


@enum.unique
class CertificateType(enum.IntEnum):
    # pre-certificate types
    NCP_W_NATURAL_PERSON_PRE_CERTIFICATE = auto()
    NCP_W_LEGAL_PERSON_PRE_CERTIFICATE = auto()
    DVCP_PRE_CERTIFICATE = auto()
    IVCP_PRE_CERTIFICATE = auto()
    OVCP_PRE_CERTIFICATE = auto()
    EVCP_PRE_CERTIFICATE = auto()
    QEVCP_W_EIDAS_PRE_CERTIFICATE = auto()
    QNCP_W_IV_EIDAS_PRE_CERTIFICATE = auto()
    QNCP_W_OV_EIDAS_PRE_CERTIFICATE = auto()
    QNCP_W_GEN_NATURAL_PERSON_EIDAS_PRE_CERTIFICATE = auto()
    QNCP_W_GEN_LEGAL_PERSON_EIDAS_PRE_CERTIFICATE = auto()
    QEVCP_W_NON_EIDAS_PRE_CERTIFICATE = auto()
    QNCP_W_IV_NON_EIDAS_PRE_CERTIFICATE = auto()
    QNCP_W_OV_NON_EIDAS_PRE_CERTIFICATE = auto()
    QNCP_W_GEN_NATURAL_PERSON_NON_EIDAS_PRE_CERTIFICATE = auto()
    QNCP_W_GEN_LEGAL_PERSON_NON_EIDAS_PRE_CERTIFICATE = auto()
    QEVCP_W_PSD2_EIDAS_PRE_CERTIFICATE = auto()
    QEVCP_W_PSD2_EIDAS_NON_BROWSER_PRE_CERTIFICATE = auto()

    # final certificate types
    NCP_W_NATURAL_PERSON_FINAL_CERTIFICATE = auto()
    NCP_W_LEGAL_PERSON_FINAL_CERTIFICATE = auto()
    DVCP_FINAL_CERTIFICATE = auto()
    IVCP_FINAL_CERTIFICATE = auto()
    OVCP_FINAL_CERTIFICATE = auto()
    EVCP_FINAL_CERTIFICATE = auto()
    QEVCP_W_EIDAS_FINAL_CERTIFICATE = auto()
    QNCP_W_IV_EIDAS_FINAL_CERTIFICATE = auto()
    QNCP_W_OV_EIDAS_FINAL_CERTIFICATE = auto()
    QNCP_W_GEN_NATURAL_PERSON_EIDAS_FINAL_CERTIFICATE = auto()
    QNCP_W_GEN_LEGAL_PERSON_EIDAS_FINAL_CERTIFICATE = auto()
    QEVCP_W_NON_EIDAS_FINAL_CERTIFICATE = auto()
    QNCP_W_IV_NON_EIDAS_FINAL_CERTIFICATE = auto()
    QNCP_W_OV_NON_EIDAS_FINAL_CERTIFICATE = auto()
    QNCP_W_GEN_NATURAL_PERSON_NON_EIDAS_FINAL_CERTIFICATE = auto()
    QNCP_W_GEN_LEGAL_PERSON_NON_EIDAS_FINAL_CERTIFICATE = auto()
    QEVCP_W_PSD2_EIDAS_FINAL_CERTIFICATE = auto()
    QEVCP_W_PSD2_EIDAS_NON_BROWSER_FINAL_CERTIFICATE = auto()

    NCP_NATURAL_PERSON_CERTIFICATE = auto()
    NCP_LEGAL_PERSON_CERTIFICATE = auto()

    def __str__(self):
        return self.name

    @property
    def to_option_str(self):
        return self.name.replace("_", "-")

    @staticmethod
    def from_option_str(value):
        value = value.replace("-", "_").upper()

        return CertificateType[value]


DVCP_CERTIFICATE_TYPES = {
    CertificateType.DVCP_PRE_CERTIFICATE,
    CertificateType.DVCP_FINAL_CERTIFICATE,
}
IVCP_CERTIFICATE_TYPES = {
    CertificateType.IVCP_PRE_CERTIFICATE,
    CertificateType.IVCP_FINAL_CERTIFICATE,
}
OVCP_CERTIFICATE_TYPES = {
    CertificateType.OVCP_PRE_CERTIFICATE,
    CertificateType.OVCP_FINAL_CERTIFICATE,
}
EVCP_CERTIFICATE_TYPES = {
    CertificateType.EVCP_PRE_CERTIFICATE,
    CertificateType.EVCP_FINAL_CERTIFICATE,
}

QNCP_W_IV_EIDAS_CERTIFICATE_TYPES = {
    CertificateType.QNCP_W_IV_EIDAS_PRE_CERTIFICATE,
    CertificateType.QNCP_W_IV_EIDAS_FINAL_CERTIFICATE,
}
QNCP_W_IV_NON_EIDAS_CERTIFICATE_TYPES = {
    CertificateType.QNCP_W_IV_NON_EIDAS_PRE_CERTIFICATE,
    CertificateType.QNCP_W_IV_NON_EIDAS_FINAL_CERTIFICATE,
}

QNCP_W_OV_EIDAS_CERTIFICATE_TYPES = {
    CertificateType.QNCP_W_OV_EIDAS_PRE_CERTIFICATE,
    CertificateType.QNCP_W_OV_EIDAS_FINAL_CERTIFICATE,
}
QNCP_W_OV_NON_EIDAS_CERTIFICATE_TYPES = {
    CertificateType.QNCP_W_OV_NON_EIDAS_PRE_CERTIFICATE,
    CertificateType.QNCP_W_OV_NON_EIDAS_FINAL_CERTIFICATE,
}

QNCP_W_IV_CERTIFICATE_TYPES = (
    QNCP_W_IV_EIDAS_CERTIFICATE_TYPES | QNCP_W_IV_NON_EIDAS_CERTIFICATE_TYPES
)
QNCP_W_OV_CERTIFICATE_TYPES = (
    QNCP_W_OV_EIDAS_CERTIFICATE_TYPES | QNCP_W_OV_NON_EIDAS_CERTIFICATE_TYPES
)

QNCP_W_GEN_NP_EIDAS_CERTIFICATE_TYPES = {
    CertificateType.QNCP_W_GEN_NATURAL_PERSON_EIDAS_PRE_CERTIFICATE,
    CertificateType.QNCP_W_GEN_NATURAL_PERSON_EIDAS_FINAL_CERTIFICATE,
}
QNCP_W_GEN_NP_NON_EIDAS_CERTIFICATE_TYPES = {
    CertificateType.QNCP_W_GEN_NATURAL_PERSON_NON_EIDAS_PRE_CERTIFICATE,
    CertificateType.QNCP_W_GEN_NATURAL_PERSON_NON_EIDAS_FINAL_CERTIFICATE,
}

QNCP_W_GEN_NP_CERTIFICATE_TYPES = (
    QNCP_W_GEN_NP_EIDAS_CERTIFICATE_TYPES | QNCP_W_GEN_NP_NON_EIDAS_CERTIFICATE_TYPES
)

QNCP_W_GEN_LP_EIDAS_CERTIFICATE_TYPES = {
    CertificateType.QNCP_W_GEN_LEGAL_PERSON_EIDAS_PRE_CERTIFICATE,
    CertificateType.QNCP_W_GEN_LEGAL_PERSON_EIDAS_FINAL_CERTIFICATE,
}
QNCP_W_GEN_LP_NON_EIDAS_CERTIFICATE_TYPES = {
    CertificateType.QNCP_W_GEN_LEGAL_PERSON_NON_EIDAS_PRE_CERTIFICATE,
    CertificateType.QNCP_W_GEN_LEGAL_PERSON_NON_EIDAS_FINAL_CERTIFICATE,
}

QNCP_W_GEN_LP_CERTIFICATE_TYPES = (
    QNCP_W_GEN_LP_EIDAS_CERTIFICATE_TYPES | QNCP_W_GEN_LP_NON_EIDAS_CERTIFICATE_TYPES
)

QNCP_W_GEN_CERTIFICATE_TYPES = (
    QNCP_W_GEN_NP_CERTIFICATE_TYPES | QNCP_W_GEN_LP_CERTIFICATE_TYPES
)

NCP_W_NATURAL_PERSON_CERTIFICATE_TYPES = {
    CertificateType.NCP_W_NATURAL_PERSON_PRE_CERTIFICATE,
    CertificateType.NCP_W_NATURAL_PERSON_FINAL_CERTIFICATE,
} | QNCP_W_GEN_NP_CERTIFICATE_TYPES

NCP_W_LEGAL_PERSON_CERTIFICATE_TYPES = {
    CertificateType.NCP_W_LEGAL_PERSON_PRE_CERTIFICATE,
    CertificateType.NCP_W_LEGAL_PERSON_FINAL_CERTIFICATE,
} | QNCP_W_GEN_LP_CERTIFICATE_TYPES

NCP_W_CERTIFICATE_TYPES = (
    NCP_W_NATURAL_PERSON_CERTIFICATE_TYPES | NCP_W_LEGAL_PERSON_CERTIFICATE_TYPES
)

NCP_NP_CERTIFICATE_TYPES = {
    CertificateType.NCP_NATURAL_PERSON_CERTIFICATE
} | NCP_W_NATURAL_PERSON_CERTIFICATE_TYPES

NCP_LP_CERTIFICATE_TYPES = {
    CertificateType.NCP_LEGAL_PERSON_CERTIFICATE
} | NCP_W_LEGAL_PERSON_CERTIFICATE_TYPES

QEVCP_W_PSD2_EIDAS_NON_BROWSER_CERTIFICATE_TYPES = {
    CertificateType.QEVCP_W_PSD2_EIDAS_NON_BROWSER_PRE_CERTIFICATE,
    CertificateType.QEVCP_W_PSD2_EIDAS_NON_BROWSER_FINAL_CERTIFICATE,
}

QEVCP_W_PSD2_EIDAS_CERTIFICATE_TYPES = {
    CertificateType.QEVCP_W_PSD2_EIDAS_PRE_CERTIFICATE,
    CertificateType.QEVCP_W_PSD2_EIDAS_FINAL_CERTIFICATE,
} | QEVCP_W_PSD2_EIDAS_NON_BROWSER_CERTIFICATE_TYPES

PSD2_EIDAS_CERTIFICATE_TYPES = QEVCP_W_PSD2_EIDAS_CERTIFICATE_TYPES

QEVCP_W_EIDAS_CERTIFICATE_TYPES = {
    CertificateType.QEVCP_W_EIDAS_PRE_CERTIFICATE,
    CertificateType.QEVCP_W_EIDAS_FINAL_CERTIFICATE,
} | QEVCP_W_PSD2_EIDAS_CERTIFICATE_TYPES

QEVCP_W_NON_EIDAS_CERTIFICATE_TYPES = {
    CertificateType.QEVCP_W_NON_EIDAS_PRE_CERTIFICATE,
    CertificateType.QEVCP_W_NON_EIDAS_FINAL_CERTIFICATE,
}

QEVCP_W_CERTIFICATE_TYPES = (
    QEVCP_W_EIDAS_CERTIFICATE_TYPES | QEVCP_W_NON_EIDAS_CERTIFICATE_TYPES
)

QNCP_W_CERTIFICATE_TYPES = (
    QNCP_W_IV_CERTIFICATE_TYPES
    | QNCP_W_OV_CERTIFICATE_TYPES
    | QEVCP_W_CERTIFICATE_TYPES
)

CABF_DV_CERTIFICATE_TYPES = DVCP_CERTIFICATE_TYPES
CABF_IV_CERTIFICATE_TYPES = IVCP_CERTIFICATE_TYPES | QNCP_W_IV_CERTIFICATE_TYPES
CABF_OV_CERTIFICATE_TYPES = OVCP_CERTIFICATE_TYPES | QNCP_W_OV_CERTIFICATE_TYPES
CABF_EV_CERTIFICATE_TYPES = (
    EVCP_CERTIFICATE_TYPES
    | QEVCP_W_CERTIFICATE_TYPES
    | QEVCP_W_PSD2_EIDAS_CERTIFICATE_TYPES
)

CABF_CERTIFICATE_TYPES = (
    CABF_DV_CERTIFICATE_TYPES
    | CABF_IV_CERTIFICATE_TYPES
    | CABF_OV_CERTIFICATE_TYPES
    | CABF_EV_CERTIFICATE_TYPES
)

NATURAL_PERSON_CERTIFICATE_TYPES = (
    CABF_IV_CERTIFICATE_TYPES
    | QNCP_W_GEN_NP_CERTIFICATE_TYPES
    | NCP_NP_CERTIFICATE_TYPES
)

LEGAL_PERSON_CERTIFICATE_TYPES = (
    CABF_OV_CERTIFICATE_TYPES
    | CABF_EV_CERTIFICATE_TYPES
    | QNCP_W_GEN_LP_CERTIFICATE_TYPES
    | NCP_LP_CERTIFICATE_TYPES
)

QWAC_TYPES = (
    QEVCP_W_CERTIFICATE_TYPES
    | QNCP_W_IV_CERTIFICATE_TYPES
    | QNCP_W_OV_CERTIFICATE_TYPES
    | QNCP_W_GEN_NP_CERTIFICATE_TYPES
    | QNCP_W_GEN_LP_CERTIFICATE_TYPES
    | QEVCP_W_PSD2_EIDAS_CERTIFICATE_TYPES
)

WEB_AUTHENTICATION_CERTIFICATE_TYPES = (
    CABF_CERTIFICATE_TYPES | QWAC_TYPES | NCP_W_CERTIFICATE_TYPES
)

NON_CABF_WEB_AUTHENTICATION_CERTIFICATE_TYPES = (
    WEB_AUTHENTICATION_CERTIFICATE_TYPES - CABF_CERTIFICATE_TYPES
)

EU_QWAC_TYPES = {
    CertificateType.QEVCP_W_EIDAS_PRE_CERTIFICATE,
    CertificateType.QEVCP_W_EIDAS_FINAL_CERTIFICATE,
    CertificateType.QNCP_W_IV_EIDAS_PRE_CERTIFICATE,
    CertificateType.QNCP_W_IV_EIDAS_FINAL_CERTIFICATE,
    CertificateType.QNCP_W_OV_EIDAS_PRE_CERTIFICATE,
    CertificateType.QNCP_W_OV_EIDAS_FINAL_CERTIFICATE,
    CertificateType.QNCP_W_GEN_NATURAL_PERSON_EIDAS_PRE_CERTIFICATE,
    CertificateType.QNCP_W_GEN_NATURAL_PERSON_EIDAS_FINAL_CERTIFICATE,
    CertificateType.QNCP_W_GEN_LEGAL_PERSON_EIDAS_PRE_CERTIFICATE,
    CertificateType.QNCP_W_GEN_LEGAL_PERSON_EIDAS_FINAL_CERTIFICATE,
    CertificateType.QEVCP_W_PSD2_EIDAS_PRE_CERTIFICATE,
    CertificateType.QEVCP_W_PSD2_EIDAS_FINAL_CERTIFICATE,
    CertificateType.QEVCP_W_PSD2_EIDAS_NON_BROWSER_PRE_CERTIFICATE,
    CertificateType.QEVCP_W_PSD2_EIDAS_NON_BROWSER_FINAL_CERTIFICATE,
}

NON_EU_QWAC_TYPES = QWAC_TYPES - EU_QWAC_TYPES

ETSI_TYPE_TO_CABF_SERVERAUTH_TYPE_MAPPINGS = {
    CertificateType.DVCP_PRE_CERTIFICATE: serverauth_constants.CertificateType.DV_PRE_CERTIFICATE,
    CertificateType.IVCP_PRE_CERTIFICATE: serverauth_constants.CertificateType.IV_PRE_CERTIFICATE,
    CertificateType.OVCP_PRE_CERTIFICATE: serverauth_constants.CertificateType.OV_PRE_CERTIFICATE,
    CertificateType.EVCP_PRE_CERTIFICATE: serverauth_constants.CertificateType.EV_PRE_CERTIFICATE,
    CertificateType.QEVCP_W_EIDAS_PRE_CERTIFICATE: serverauth_constants.CertificateType.EV_PRE_CERTIFICATE,
    CertificateType.QNCP_W_IV_EIDAS_PRE_CERTIFICATE: serverauth_constants.CertificateType.IV_PRE_CERTIFICATE,
    CertificateType.QNCP_W_OV_EIDAS_PRE_CERTIFICATE: serverauth_constants.CertificateType.OV_PRE_CERTIFICATE,
    CertificateType.QEVCP_W_NON_EIDAS_PRE_CERTIFICATE: serverauth_constants.CertificateType.EV_PRE_CERTIFICATE,
    CertificateType.QNCP_W_IV_NON_EIDAS_PRE_CERTIFICATE: serverauth_constants.CertificateType.IV_PRE_CERTIFICATE,
    CertificateType.QNCP_W_OV_NON_EIDAS_PRE_CERTIFICATE: serverauth_constants.CertificateType.OV_PRE_CERTIFICATE,
    CertificateType.QEVCP_W_PSD2_EIDAS_PRE_CERTIFICATE: serverauth_constants.CertificateType.EV_PRE_CERTIFICATE,
    CertificateType.QEVCP_W_PSD2_EIDAS_NON_BROWSER_PRE_CERTIFICATE: (
        serverauth_constants.CertificateType.EV_PRE_CERTIFICATE
    ),
    CertificateType.DVCP_FINAL_CERTIFICATE: serverauth_constants.CertificateType.DV_FINAL_CERTIFICATE,
    CertificateType.IVCP_FINAL_CERTIFICATE: serverauth_constants.CertificateType.IV_FINAL_CERTIFICATE,
    CertificateType.OVCP_FINAL_CERTIFICATE: serverauth_constants.CertificateType.OV_FINAL_CERTIFICATE,
    CertificateType.EVCP_FINAL_CERTIFICATE: serverauth_constants.CertificateType.EV_FINAL_CERTIFICATE,
    CertificateType.QEVCP_W_EIDAS_FINAL_CERTIFICATE: serverauth_constants.CertificateType.EV_FINAL_CERTIFICATE,
    CertificateType.QNCP_W_IV_EIDAS_FINAL_CERTIFICATE: serverauth_constants.CertificateType.IV_FINAL_CERTIFICATE,
    CertificateType.QNCP_W_OV_EIDAS_FINAL_CERTIFICATE: serverauth_constants.CertificateType.OV_FINAL_CERTIFICATE,
    CertificateType.QEVCP_W_NON_EIDAS_FINAL_CERTIFICATE: serverauth_constants.CertificateType.EV_FINAL_CERTIFICATE,
    CertificateType.QNCP_W_IV_NON_EIDAS_FINAL_CERTIFICATE: serverauth_constants.CertificateType.IV_FINAL_CERTIFICATE,
    CertificateType.QNCP_W_OV_NON_EIDAS_FINAL_CERTIFICATE: serverauth_constants.CertificateType.OV_FINAL_CERTIFICATE,
    CertificateType.QEVCP_W_PSD2_EIDAS_FINAL_CERTIFICATE: serverauth_constants.CertificateType.EV_FINAL_CERTIFICATE,
    CertificateType.QEVCP_W_PSD2_EIDAS_NON_BROWSER_FINAL_CERTIFICATE: (
        serverauth_constants.CertificateType.EV_FINAL_CERTIFICATE
    ),
}



import datetime

from pyasn1_alt_modules import rfc5280

from pkilint import validation, document


class LegacyGenerationSunsetValidator(validation.Validator):
    """
    SMBR, section 7.1.6.1:
    Effective July 15, 2025 S/MIME Subscriber Certificates SHALL NOT be issued using the Legacy Generation profiles
    2.23.140.1.5.1.1, 2.23.140.1.5.2.1, 2.23.140.1.5.3.1, or 2.23.140.1.5.4.1.
    """

    VALIDATION_LEGACY_GENERATION_CERTIFICATE_ISSUED_AFTER_PROHIBITION = (
        validation.ValidationFinding(
            validation.ValidationFindingSeverity.ERROR,
            "cabf.smime.legacy_generation_certificate_issued_after_prohibition",
        )
    )

    _LEGACY_GENERATION_SUNSET_DATE = datetime.datetime(
        2025, 7, 15, 0, 0, 0, tzinfo=datetime.timezone.utc
    )

    def __init__(
        self, validity_period_start_retriever: document.ValidityPeriodStartRetriever
    ):
        super().__init__(
            validations=[
                self.VALIDATION_LEGACY_GENERATION_CERTIFICATE_ISSUED_AFTER_PROHIBITION
            ],
            pdu_class=rfc5280.Validity,
        )

        self._validity_period_start_retriever = validity_period_start_retriever

    def validate(self, node):
        if (
            self._validity_period_start_retriever(node.document)
            >= self._LEGACY_GENERATION_SUNSET_DATE
        ):
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_LEGACY_GENERATION_CERTIFICATE_ISSUED_AFTER_PROHIBITION
            )



import typing
from typing import List

from pyasn1_alt_modules import rfc5280, rfc6962, rfc3739

from pkilint import validation, finding_filter, cabf, document
from pkilint.cabf import serverauth
from pkilint.cabf.serverauth import (
    serverauth_constants,
    serverauth_name,
    serverauth_finding_filter,
)
from pkilint.common import organization_id, alternative_name
from pkilint.etsi import (
    etsi_constants,
    ts_119_495,
    en_319_412_5,
    en_319_412_1,
    en_319_412_2,
    en_319_412_3,
    ts_119_312,
    en_319_412_4,
    etsi_shared,
    etsi_finding_filter,
    en_319_411_1,
)
from pkilint.etsi.asn1 import (
    en_319_412_1 as en_319_412_asn1,
    en_319_412_5 as en_319_412_5_asn1,
    ts_119_495 as ts_119_495_asn1,
)
from pkilint.etsi.etsi_constants import CertificateType
from pkilint.itu import x520_name_unbounded
from pkilint.pkix import certificate


def determine_certificate_type(cert: certificate.RFC5280Certificate) -> CertificateType:
    qualified_statement_ids = cert.qualified_statement_ids
    policy_oids = cert.policy_oids

    is_qualified = en_319_412_5_asn1.id_etsi_qcs_QcCompliance in qualified_statement_ids
    is_eidas_qualified = (
        is_qualified
        and en_319_412_5_asn1.id_etsi_qcs_QcCClegislation not in qualified_statement_ids
    )
    is_precert = cert.get_extension_by_oid(rfc6962.id_ce_criticalPoison) is not None
    is_webauth = rfc5280.id_kp_serverAuth in cert.extended_key_usages

    if serverauth_constants.ID_POLICY_EV in policy_oids:
        is_psd2 = ts_119_495_asn1.id_etsi_psd2_qcStatement in qualified_statement_ids

        if is_psd2:
            return (
                CertificateType.QEVCP_W_PSD2_EIDAS_PRE_CERTIFICATE
                if is_precert
                else CertificateType.QEVCP_W_PSD2_EIDAS_FINAL_CERTIFICATE
            )
        elif is_eidas_qualified:
            return (
                CertificateType.QEVCP_W_EIDAS_PRE_CERTIFICATE
                if is_precert
                else CertificateType.QEVCP_W_EIDAS_FINAL_CERTIFICATE
            )
        elif is_qualified:
            return (
                CertificateType.QEVCP_W_NON_EIDAS_PRE_CERTIFICATE
                if is_precert
                else CertificateType.QEVCP_W_NON_EIDAS_FINAL_CERTIFICATE
            )
        else:
            return (
                CertificateType.EVCP_PRE_CERTIFICATE
                if is_precert
                else CertificateType.EVCP_FINAL_CERTIFICATE
            )
    elif serverauth_constants.ID_POLICY_OV in policy_oids:
        if is_eidas_qualified:
            return (
                CertificateType.QNCP_W_OV_EIDAS_PRE_CERTIFICATE
                if is_precert
                else CertificateType.QNCP_W_OV_EIDAS_FINAL_CERTIFICATE
            )
        elif is_qualified:
            return (
                CertificateType.QNCP_W_OV_NON_EIDAS_PRE_CERTIFICATE
                if is_precert
                else CertificateType.QNCP_W_OV_NON_EIDAS_FINAL_CERTIFICATE
            )
        else:
            return (
                CertificateType.OVCP_PRE_CERTIFICATE
                if is_precert
                else CertificateType.OVCP_FINAL_CERTIFICATE
            )
    elif serverauth_constants.ID_POLICY_IV in policy_oids:
        if is_eidas_qualified:
            return (
                CertificateType.QNCP_W_IV_EIDAS_PRE_CERTIFICATE
                if is_precert
                else CertificateType.QNCP_W_IV_EIDAS_FINAL_CERTIFICATE
            )
        elif is_qualified:
            return (
                CertificateType.QNCP_W_IV_NON_EIDAS_PRE_CERTIFICATE
                if is_precert
                else CertificateType.QNCP_W_IV_NON_EIDAS_FINAL_CERTIFICATE
            )
        else:
            return (
                CertificateType.IVCP_PRE_CERTIFICATE
                if is_precert
                else CertificateType.IVCP_FINAL_CERTIFICATE
            )
    elif serverauth_constants.ID_POLICY_DV in policy_oids:
        return (
            CertificateType.DVCP_PRE_CERTIFICATE
            if is_precert
            else CertificateType.DVCP_FINAL_CERTIFICATE
        )
    else:
        is_natural_person = any(
            (
                cert.get_subject_attributes_by_type(rfc5280.id_at_givenName),
                cert.get_subject_attributes_by_type(rfc5280.id_at_surname),
                cert.get_subject_attributes_by_type(rfc5280.id_at_pseudonym),
            )
        )

        if is_natural_person:
            if is_webauth:
                if is_eidas_qualified:
                    return (
                        CertificateType.QNCP_W_GEN_NATURAL_PERSON_EIDAS_PRE_CERTIFICATE
                        if is_precert
                        else CertificateType.QNCP_W_GEN_NATURAL_PERSON_EIDAS_FINAL_CERTIFICATE
                    )
                elif is_qualified:
                    return (
                        CertificateType.QNCP_W_GEN_NATURAL_PERSON_NON_EIDAS_PRE_CERTIFICATE
                        if is_precert
                        else CertificateType.QNCP_W_GEN_NATURAL_PERSON_NON_EIDAS_FINAL_CERTIFICATE
                    )
                else:
                    return (
                        CertificateType.NCP_W_NATURAL_PERSON_PRE_CERTIFICATE
                        if is_precert
                        else CertificateType.NCP_W_NATURAL_PERSON_FINAL_CERTIFICATE
                    )

            return CertificateType.NCP_NATURAL_PERSON_CERTIFICATE
        else:
            if is_webauth:
                if is_eidas_qualified:
                    return (
                        CertificateType.QNCP_W_GEN_LEGAL_PERSON_EIDAS_PRE_CERTIFICATE
                        if is_precert
                        else CertificateType.QNCP_W_GEN_LEGAL_PERSON_EIDAS_FINAL_CERTIFICATE
                    )
                elif is_qualified:
                    return (
                        CertificateType.QNCP_W_GEN_LEGAL_PERSON_NON_EIDAS_PRE_CERTIFICATE
                        if is_precert
                        else CertificateType.QNCP_W_GEN_LEGAL_PERSON_NON_EIDAS_FINAL_CERTIFICATE
                    )
                else:
                    return (
                        CertificateType.NCP_W_LEGAL_PERSON_PRE_CERTIFICATE
                        if is_precert
                        else CertificateType.NCP_W_LEGAL_PERSON_FINAL_CERTIFICATE
                    )

            return CertificateType.NCP_LEGAL_PERSON_CERTIFICATE


def create_decoding_validators(
    certificate_type: CertificateType,
) -> List[validation.Validator]:
    if certificate_type in etsi_constants.CABF_CERTIFICATE_TYPES:
        return serverauth.create_decoding_validators()
    else:
        name_attribute_mappings = cabf.NAME_ATTRIBUTE_MAPPINGS.copy()

        """
        From EN 319 412-2:
        NAT-4.2.4-18 If present, the size of givenName, surname, pseudonym, commonName, organizationName and
        organizationalUnitName may be longer than the limit as stated in IETF RFC 5280 [1].
        
        From EN 319 412-3:
        LEG-4.2.1-9: If present, the size of organizationName, organizationalUnitName and commonName may
        be longer than the limit as stated in IETF RFC 5280 [3].
        
        From EN 319 412-4:
        WEB-4.1.3-4: The following certificate profile requirements specified in the BRG [9] shall apply for subject
        certificate fields addressed by the following sub-sections of BRG:
        ...
        (c) 7.1.4.2.2 Subject Distinguished Name - commonName.
        """
        name_attribute_mappings.update(
            {
                k: v
                for k, v in x520_name_unbounded.UNBOUNDED_ATTRIBUTE_TYPE_MAPPINGS.items()
                if k
                in (
                    rfc5280.id_at_organizationName,
                    rfc5280.id_at_organizationalUnitName,
                    rfc5280.id_at_pseudonym,
                )
            }
        )

        if certificate_type not in etsi_constants.WEB_AUTHENTICATION_CERTIFICATE_TYPES:
            name_attribute_mappings[
                rfc5280.id_at_commonName
            ]: x520_name_unbounded.X520CommonNameUnbounded()

        additional_validators = [
            certificate.create_qc_statements_decoder(asn1.ETSI_QC_STATEMENTS_MAPPINGS)
        ]

        return certificate.create_decoding_validators(
            name_attribute_mappings, cabf.EXTENSION_MAPPINGS, additional_validators
        )


def create_validators(
    certificate_type: CertificateType,
    validity_period_start_retriever: typing.Optional[
        document.ValidityPeriodStartRetriever
    ] = None,
    additional_validity_validators=None,
    additional_spki_validators=None,
    additional_name_validators=None,
    additional_extension_validators=None,
    additional_top_level_validators=None,
) -> List[validation.Validator]:
    subject_validators = [
        en_319_412_1.LegalPersonOrganizationIdentifierValidator(),
        en_319_412_1.NaturalPersonIdentifierValidator(),
        en_319_412_1.EidasLegalPersonIdentifierValidator(),
        en_319_412_1.NaturalPersonEidasIdentifierValidator(),
        organization_id.OrganizationIdentifierLeiValidator(),
        en_319_412_3.LegalPersonOrganizationAttributesEqualityValidator(),
    ]

    if additional_name_validators:
        subject_validators.extend(additional_name_validators)

    qc_statement_validators = [
        ts_119_495.RolesOfPspValidator(),
        ts_119_495.NCANameLatinCharactersValidator(),
        ts_119_495.NCAIdValidator(),
        en_319_412_5.QcCCLegislationCountryCodeValidator(),
        en_319_412_5.QcEuRetentionPeriodValidator(),
        en_319_412_5.QcTypeValidator(certificate_type),
        en_319_412_5.QcEuPDSHttpsURLValidator(),
        en_319_412_5.QcEuLimitValueValidator(),
        en_319_412_5.QcEuPDSLanguageValidator(),
        en_319_412_1.LegalPersonIdentifierNameRegistrationAuthoritiesValidator(),
        en_319_412_1.NaturalPersonIdentifierNameRegistrationAuthoritiesValidator(),
        en_319_412_5.QcStatementIdentifierAllowanceValidator(certificate_type),
    ]

    qc_statements_validator_container = validation.ValidatorContainer(
        validators=qc_statement_validators, pdu_class=rfc3739.QCStatements
    )

    extension_validators = [
        en_319_412_2.QualifiedCertificatePoliciesValidator(certificate_type),
        en_319_412_5.QcStatementsExtensionCriticalityValidator(),
        ts_119_495.Psd2CertificatePolicyOidPresenceValidator(certificate_type),
        qc_statements_validator_container,
    ]

    if additional_extension_validators:
        extension_validators.extend(additional_extension_validators)

    spki_validators = [
        ts_119_312.RsaKeyValidator(),
        ts_119_312.AllowedPublicKeyTypeValidator(),
    ]

    if additional_spki_validators:
        spki_validators.extend(additional_spki_validators)

    top_level_validators = [
        en_319_412_2.ExtensionsPresenceValidator(),
        ts_119_312.AllowedSignatureAlgorithmValidator(
            path="certificate.tbsCertificate.signature"
        ),
    ]

    if additional_top_level_validators:
        top_level_validators.extend(additional_top_level_validators)

    if (
        certificate_type in etsi_constants.LEGAL_PERSON_CERTIFICATE_TYPES
        and certificate_type not in etsi_constants.CABF_CERTIFICATE_TYPES
    ):
        subject_validators.extend(
            [
                en_319_412_3.LegalPersonSubjectAttributeAllowanceValidator(),
                en_319_412_3.LegalPersonDuplicateAttributeAllowanceValidator(),
                en_319_412_3.LegalPersonCountryCodeValidator(),
            ]
        )

    elif (
        certificate_type in etsi_constants.NATURAL_PERSON_CERTIFICATE_TYPES
        and certificate_type not in etsi_constants.CABF_CERTIFICATE_TYPES
    ):
        subject_validators.append(
            en_319_412_2.NaturalPersonSubjectAttributeAllowanceValidator()
        )

    if certificate_type not in etsi_constants.CABF_CERTIFICATE_TYPES:
        extension_validators.extend(
            [
                en_319_412_2.CertificatePoliciesCriticalityValidator(),
                en_319_412_2.SubjectAlternativeNameCriticalityValidator(),
                en_319_412_2.IssuerAlternativeNameCriticalityValidator(),
                en_319_412_2.ExtendedKeyUsageCriticalityValidator(),
                en_319_412_2.CRLDistributionPointsCriticalityValidator(),
                en_319_412_2.NaturalPersonExtensionIdentifierAllowanceValidator(
                    certificate_type
                ),
                en_319_412_2.CrlDistributionPointsExtensionPresenceValidator(),
                en_319_412_2.CrlDistributionPointsValidator(),
                en_319_412_2.AuthorityInformationAccessValidator(),
            ]
        )

        if certificate_type in etsi_constants.WEB_AUTHENTICATION_CERTIFICATE_TYPES:
            extension_validators.extend(
                [
                    en_319_412_4.NcpWExtendedKeyUsagePresenceValidator(),
                    serverauth.serverauth_subscriber.SubscriberEkuAllowanceValidator(),
                    en_319_412_4.NcpWCriticalityExtendedKeyUsageValidator(),
                    serverauth.serverauth_subscriber.SubscriberSanGeneralNameTypeValidator(),
                    serverauth_name.DnsNameLdhLabelSyntaxValidator(),
                    en_319_412_4.NcpWSubjectAltNamePresenceValidator(),
                ]
            )

        # TODO: fix commitment types when adding support for eSeal and eSignature
        if certificate_type in etsi_constants.LEGAL_PERSON_CERTIFICATE_TYPES:
            extension_validators.append(
                en_319_412_3.LegalPersonKeyUsageValidator(
                    is_content_commitment_type=None
                )
            )
        elif certificate_type in etsi_constants.NATURAL_PERSON_CERTIFICATE_TYPES:
            extension_validators.append(
                en_319_412_2.NaturalPersonKeyUsageValidator(
                    is_content_commitment_type=None
                )
            )

    if certificate_type in etsi_constants.QEVCP_W_PSD2_EIDAS_CERTIFICATE_TYPES:
        qc_statement_validators.append(ts_119_495.PresenceofQCEUPDSStatementValidator())

        subject_validators.append(ts_119_495.PsdOrganizationIdentifierFormatValidator())

    if certificate_type in etsi_constants.QNCP_W_CERTIFICATE_TYPES:
        subject_validators.append(en_319_412_4.QncpWCommonNameValidator())
    elif certificate_type in etsi_constants.NCP_W_CERTIFICATE_TYPES:
        subject_validators.append(en_319_412_4.NcpWCommonNameValidator())

    if certificate_type in etsi_constants.CABF_CERTIFICATE_TYPES:
        extension_validators.append(
            en_319_411_1.CertificatePoliciesValidator(certificate_type)
        )

        serverauth_cert_type = (
            etsi_constants.ETSI_TYPE_TO_CABF_SERVERAUTH_TYPE_MAPPINGS[certificate_type]
        )

        return serverauth.create_validators(
            serverauth_cert_type,
            validity_period_start_retriever=validity_period_start_retriever,
            additional_top_level_validators=top_level_validators,
            additional_validity_validators=additional_validity_validators,
            additional_name_validators=subject_validators,
            additional_extension_validators=extension_validators,
            additional_spki_validators=spki_validators,
        )
    else:
        spki_validator_container = validation.ValidatorContainer(
            validators=spki_validators,
            path="certificate.tbsCertificate.subjectPublicKeyInfo",
        )

        top_level_container = validation.ValidatorContainer(
            validators=top_level_validators, pdu_class=rfc5280.Certificate
        )

        extension_validators.extend(
            [
                alternative_name.create_internal_name_validator_container(
                    etsi_shared.VALIDATION_INTERNAL_DOMAIN_NAME,
                    etsi_shared.VALIDATION_INTERNAL_IP_ADDRESS,
                    allow_onion_tld=False,
                ),
                alternative_name.create_cpsuri_internal_domain_name_validator(
                    etsi_shared.VALIDATION_INTERNAL_DOMAIN_NAME
                ),
            ]
        )

        return [
            certificate.create_issuer_validator_container([]),
            certificate.create_validity_validator_container(
                additional_validity_validators
            ),
            certificate.create_subject_validator_container(subject_validators),
            certificate.create_extensions_validator_container(extension_validators),
            spki_validator_container,
            top_level_container,
        ]


def create_etsi_finding_filters(
    certificate_type,
) -> List[finding_filter.FindingDescriptionFilter]:
    if certificate_type in etsi_constants.CABF_CERTIFICATE_TYPES:
        serverauth_cert_type = (
            etsi_constants.ETSI_TYPE_TO_CABF_SERVERAUTH_TYPE_MAPPINGS[certificate_type]
        )

        filters = serverauth.create_serverauth_finding_filters(serverauth_cert_type)
    else:
        filters = [
            serverauth_finding_filter.DnsNameGeneralNamePreferredNameSyntaxFilter(),
        ]

    if (
        certificate_type
        in etsi_constants.QEVCP_W_PSD2_EIDAS_NON_BROWSER_CERTIFICATE_TYPES
    ):
        filters.append(etsi_finding_filter.Psd2CabfServerauthValidityPeriodFilter())

    return filters



from pyasn1_alt_modules import rfc5280

from pkilint import validation, oid
from pkilint.etsi import etsi_constants
from pkilint.etsi.asn1 import en_319_411_1


class CertificatePoliciesValidator(validation.Validator):
    """
    GEN-6.3.3-12: The CP identifier shall be [CHOICE]:
    ...
    """

    VALIDATION_MULTIPLE_RESERVED_POLICY_OIDS_PRESENT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "etsi.en_319_411_1.gen-6.3.3-12.multiple_reserved_policy_oids_present",
    )

    VALIDATION_PROHIBITED_RESERVED_POLICY_OID_PRESENT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "etsi.en_319_411_1.gen-6.3.3-12.prohibited_reserved_policy_oid_present",
    )

    # mapping of certificate types to ETSI policy OIDs
    _CERTIFICATE_TYPE_SET_TO_POLICY_OID_MAPPINGS = [
        (etsi_constants.CABF_EV_CERTIFICATE_TYPES, en_319_411_1.id_evcp),
        (etsi_constants.CABF_DV_CERTIFICATE_TYPES, en_319_411_1.id_dvcp),
        (etsi_constants.CABF_OV_CERTIFICATE_TYPES, en_319_411_1.id_ovcp),
        (etsi_constants.CABF_IV_CERTIFICATE_TYPES, en_319_411_1.id_ivcp),
    ]

    def __init__(self, certificate_type):
        super().__init__(
            validations=[
                self.VALIDATION_MULTIPLE_RESERVED_POLICY_OIDS_PRESENT,
                self.VALIDATION_PROHIBITED_RESERVED_POLICY_OID_PRESENT,
            ],
            pdu_class=rfc5280.CertificatePolicies,
        )

        self._expected_policy_oid = next(
            p
            for t, p in self._CERTIFICATE_TYPE_SET_TO_POLICY_OID_MAPPINGS
            if certificate_type in t
        )

    def validate(self, node):
        # extract ETSI reserved policy OIDs from certificate policy OIDs
        etsi_policy_oids = en_319_411_1.POLICY_OIDS & node.document.policy_oids

        # if multiple ETSI policy OIDs are present, then report
        if len(etsi_policy_oids) > 1:
            oids = oid.format_oids(etsi_policy_oids)

            raise validation.ValidationFindingEncountered(
                self.VALIDATION_MULTIPLE_RESERVED_POLICY_OIDS_PRESENT,
                f"Multiple reserved certificate policy OIDs present: {oids}",
            )

        # if there is a mismatch between the certificate type and reserved ETSI policy OID, then report
        if etsi_policy_oids and self._expected_policy_oid not in etsi_policy_oids:
            prohibited_oid = next(iter(etsi_policy_oids))

            raise validation.ValidationFindingEncountered(
                self.VALIDATION_PROHIBITED_RESERVED_POLICY_OID_PRESENT,
                f"Prohibited reserved certificate policy OID present: {prohibited_oid}",
            )



import re

from pyasn1.type import univ, char, constraint

from pkilint import validation

ISO_OID_ARC = univ.ObjectIdentifier("1.3.6.1.4.1.52266")


id_ce_lei = ISO_OID_ARC + (1,)
id_ce_role = ISO_OID_ARC + (2,)

ub_leiRole_length = 100


class Lei(char.PrintableString):
    subtypeSpec = constraint.ValueSizeConstraint(20, 20)


class Role(char.PrintableString):
    subtypeSpec = constraint.ValueSizeConstraint(1, ub_leiRole_length)


EXTENSION_MAPPINGS = {
    id_ce_lei: Lei(),
    id_ce_role: Role(),
}


_LEI_FORMAT_RE = re.compile(r"^(?P<value>[0-9A-Z]{18})(?P<checksum>\d{2})$")


VALIDATION_INVALID_LEI_FORMAT = validation.ValidationFinding(
    validation.ValidationFindingSeverity.ERROR, "iso.lei.invalid_lei_format"
)

VALIDATION_INVALID_LEI_CHECKSUM = validation.ValidationFinding(
    validation.ValidationFindingSeverity.ERROR, "iso.lei.invalid_lei_checksum"
)


def _calculate_checksum(value):
    subs = {chr(l): str(10 + l - ord("A")) for l in range(ord("A"), ord("Z") + 1)}

    for l, n in subs.items():
        value = value.replace(l, n)

    value += "00"

    value_int = int(value)

    remainder = value_int % 97

    return 98 - remainder


def validate_lei(lei: str):
    m = _LEI_FORMAT_RE.match(lei)

    if m is None:
        raise validation.ValidationFindingEncountered(
            VALIDATION_INVALID_LEI_FORMAT, f'Invalid LEI format: "{lei}"'
        )

    value_part = m.group("value")
    actual_checksum = int(m.group("checksum"))

    expected_checksum = _calculate_checksum(value_part)

    if actual_checksum != expected_checksum:
        raise validation.ValidationFindingEncountered(
            VALIDATION_INVALID_LEI_CHECKSUM,
            f"Expected: {expected_checksum}, actual: {actual_checksum}",
        )


class LeiExtensionValueSyntaxValidator(validation.Validator):
    def __init__(self):
        super().__init__(
            validations=[
                VALIDATION_INVALID_LEI_FORMAT,
                VALIDATION_INVALID_LEI_CHECKSUM,
            ],
            pdu_class=Lei,
        )

    def validate(self, node):
        value = str(node.pdu)

        validate_lei(value)



from pyasn1_alt_modules import rfc5280

from pkilint import validation


class CorrectVersionValidator(validation.ScalarFieldValueEqualityValidator):
    def __init__(self):
        super().__init__(
            path="certificate.tbsCertificate.version",
            value=rfc5280.Version.namedValues["v3"],
            validations=validation.ValidationFinding(
                validation.ValidationFindingSeverity.ERROR,
                "pkix.certificate_version_is_not_v3",
            ),
        )


class SignatureAlgorithmMatchValidator(validation.DEREqualityValidator):
    def __init__(self):
        super().__init__(
            other_node_retriever=(lambda n: n.navigate("^.tbsCertificate.signature")),
            path="certificate.signatureAlgorithm",
            validation=validation.ValidationFinding(
                validation.ValidationFindingSeverity.ERROR,
                "pkix.certificate_signature_algorithm_mismatch",
            ),
        )


class IssuerUniqueIdAbsenceValidator(validation.NodePresenceValidator):
    VALIDATION_ISSUER_UNIQUE_ID_PRESENT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR, "pkix.issuer_unique_id_present"
    )

    def __init__(self):
        super().__init__(
            node_retriever=lambda n: n.navigate("issuerUniqueID"),
            presence_finding=self.VALIDATION_ISSUER_UNIQUE_ID_PRESENT,
            pdu_class=rfc5280.TBSCertificate,
        )


class SubjectUniqueIdAbsenceValidator(validation.NodePresenceValidator):
    VALIDATION_SUBJECT_UNIQUE_ID_PRESENT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR, "pkix.subject_unique_id_present"
    )

    def __init__(self):
        super().__init__(
            node_retriever=lambda n: n.navigate("subjectUniqueID"),
            presence_finding=self.VALIDATION_SUBJECT_UNIQUE_ID_PRESENT,
            pdu_class=rfc5280.TBSCertificate,
        )



import datetime

from pkilint import validation, document
from pkilint.pkix import time


class CertificateSaneValidityPeriodValidator(time.SaneValidityPeriodValidator):
    VALIDATION_NEGATIVE_VALIDITY_PERIOD = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "pkix.certificate_negative_validity_period",
    )

    def __init__(self):
        super().__init__(
            end_validity_node_retriever=lambda n: n.navigate("^.notAfter"),
            path="certificate.tbsCertificate.validity.notBefore",
            validation=self.VALIDATION_NEGATIVE_VALIDITY_PERIOD,
        )


class CertificateValidityPeriodStartRetriever(document.ValidityPeriodStartRetriever):
    def __call__(self, certificate, *args, **kwargs) -> datetime.datetime:
        return certificate.not_before



from urllib.parse import urlparse

from pyasn1_alt_modules import rfc5280, rfc3279, rfc5480, rfc8398, rfc8410, rfc3739

import pkilint.adobe.asn1 as adobe_asn1
from pkilint import oid, validation
from pkilint.cabf.smime import smime_constants
from pkilint.cabf.smime.smime_constants import Generation, ValidationLevel
from pkilint.cabf.smime.smime_name import get_email_addresses_from_san
from pkilint.iso import lei
from pkilint.itu import bitstring
from pkilint.pkix import extension
from pkilint.pkix.certificate.certificate_extension import KeyUsageBitName


class CertificatePoliciesPresenceValidator(extension.ExtensionPresenceValidator):
    VALIDATION_CERTIFICATE_POLICIES_EXTENSION_ABSENT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "cabf.smime.certificate_policies_extension_missing",
    )

    def __init__(self):
        super().__init__(
            extension_oid=rfc5280.id_ce_certificatePolicies,
            validation=self.VALIDATION_CERTIFICATE_POLICIES_EXTENSION_ABSENT,
            pdu_class=rfc5280.Extensions,
        )


class ExtendedKeyUsagePresenceValidator(extension.ExtensionPresenceValidator):
    VALIDATION_EKU_EXTENSION_ABSENT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "cabf.smime.extended_key_usage_extension_missing",
    )

    def __init__(self):
        super().__init__(
            extension_oid=rfc5280.id_ce_extKeyUsage,
            validation=self.VALIDATION_EKU_EXTENSION_ABSENT,
            pdu_class=rfc5280.Extensions,
        )


class CabfSmimeKeyUsagePresenceValidator(extension.ExtensionPresenceValidator):
    VALIDATION_KU_EXTENSION_ABSENT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "cabf.smime.key_usage_extension_missing",
    )

    def __init__(self):
        super().__init__(
            extension_oid=rfc5280.id_ce_keyUsage,
            validation=self.VALIDATION_KU_EXTENSION_ABSENT,
            pdu_class=rfc5280.Extensions,
        )


class SubjectAlternativeNamePresenceValidator(extension.ExtensionPresenceValidator):
    VALIDATION_SAN_EXTENSION_ABSENT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR, "cabf.smime.san_extension_missing"
    )

    def __init__(self):
        super().__init__(
            extension_oid=rfc5280.id_ce_subjectAltName,
            validation=self.VALIDATION_SAN_EXTENSION_ABSENT,
            pdu_class=rfc5280.Extensions,
        )


class LeiPresenceValidator(validation.Validator):
    VALIDATION_LEI_EXTENSION_PROHIBITED = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "cabf.smime.lei_extension_prohibited",
    )

    def __init__(self, validation_level):
        super().__init__(
            validations=[self.VALIDATION_LEI_EXTENSION_PROHIBITED], pdu_class=lei.Lei
        )

        self._validation_level = validation_level

    def validate(self, node):
        if self._validation_level in {
            ValidationLevel.MAILBOX,
            ValidationLevel.INDIVIDUAL,
        }:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_LEI_EXTENSION_PROHIBITED
            )


class LeiRolePresenceValidator(validation.Validator):
    VALIDATION_LEI_ROLE_EXTENSION_PROHIBITED = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "cabf.smime.lei_role_extension_prohibited",
    )

    def __init__(self, validation_level):
        super().__init__(
            validations=[self.VALIDATION_LEI_ROLE_EXTENSION_PROHIBITED],
            pdu_class=lei.Role,
        )

        self._validation_level = validation_level

    def validate(self, node):
        if self._validation_level != ValidationLevel.SPONSORED:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_LEI_ROLE_EXTENSION_PROHIBITED
            )


class LeiCriticalityValidator(extension.ExtensionCriticalityValidator):
    VALIDATION_LEI_EXTENSION_CRITICAL = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR, "cabf.smime.lei_extension_critical"
    )

    def __init__(self):
        super().__init__(
            type_oid=lei.id_ce_lei,
            is_critical=False,
            validation=self.VALIDATION_LEI_EXTENSION_CRITICAL,
        )


class LeiRoleCriticalityValidator(extension.ExtensionCriticalityValidator):
    VALIDATION_LEI_ROLE_EXTENSION_CRITICAL = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "cabf.smime.lei_role_extension_critical",
    )

    def __init__(self):
        super().__init__(
            type_oid=lei.id_ce_role,
            is_critical=False,
            validation=self.VALIDATION_LEI_ROLE_EXTENSION_CRITICAL,
        )


class RequiredPolicyIdentifierValidator(validation.Validator):
    VALIDATION_NO_CABF_RESERVED_OID = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "cabf.smime.no_required_reserved_policy_oid",
    )

    VALIDATION_MULTIPLE_RESERVED_OIDS = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "cabf.smime.multiple_reserved_policy_oids",
    )

    VALIDATION_ANYPOLICY_PRESENT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR, "cabf.smime.anypolicy_present"
    )

    def __init__(self, validation_level, generation):
        self._validation_level = validation_level
        self._generation = generation

        super().__init__(
            validations=[
                self.VALIDATION_NO_CABF_RESERVED_OID,
                self.VALIDATION_MULTIPLE_RESERVED_OIDS,
                self.VALIDATION_ANYPOLICY_PRESENT,
            ],
            pdu_class=rfc5280.CertificatePolicies,
        )

    def validate(self, node):
        oids = set(
            (pi.children["policyIdentifier"].pdu for pi in node.children.values())
        )

        findings = []

        if rfc5280.anyPolicy in oids:
            findings.append(
                validation.ValidationFindingDescription(
                    self.VALIDATION_ANYPOLICY_PRESENT, None
                )
            )

        expected_oid = smime_constants.get_policy_oid(
            self._validation_level, self._generation
        )

        all_smime_oids = set()
        for g in smime_constants.Generation:
            for v in smime_constants.ValidationLevel:
                all_smime_oids.add(smime_constants.get_policy_oid(v, g))

        cert_smime_oids = oids.intersection(all_smime_oids)

        if expected_oid not in cert_smime_oids:
            findings.append(
                validation.ValidationFindingDescription(
                    self.VALIDATION_NO_CABF_RESERVED_OID,
                    f"Required policy OID {str(expected_oid)} is missing",
                )
            )

        if len(cert_smime_oids) > 1:
            desc = ", ".join(map(str, cert_smime_oids))

            findings.append(
                validation.ValidationFindingDescription(
                    self.VALIDATION_MULTIPLE_RESERVED_OIDS,
                    f"Multiple CA/B Reserved Policy OIDs found: {desc}",
                )
            )

        return validation.ValidationResult(self, node, findings)


class SmimeUserNoticeValidator(validation.Validator):
    VALIDATION_USERNOTICE_HAS_NOTICEREF = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "cabf.smime.usernotice_has_noticeref",
    )

    def __init__(self):
        super().__init__(
            validations=self.VALIDATION_USERNOTICE_HAS_NOTICEREF,
            pdu_class=rfc5280.UserNotice,
        )

    def validate(self, node):
        if "noticeRef" in node.children:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_USERNOTICE_HAS_NOTICEREF
            )


class AllowedExtendedKeyUsageValidator(validation.Validator):
    VALIDATION_EMAIL_PROTECTION_EKU_MISSING = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "cabf.smime.emailprotection_eku_missing",
    )

    VALIDATION_PROHIBITED_EKU_PRESENT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR, "cabf.smime.prohibited_eku_present"
    )

    _LEGACY_MP_PROHIBITED_EKUS = {
        rfc5280.anyExtendedKeyUsage,
        rfc5280.id_kp_codeSigning,
        rfc5280.id_kp_OCSPSigning,
        rfc5280.id_kp_serverAuth,
        rfc5280.id_kp_timeStamping,
    }

    def __init__(self, generation):
        self._generation = generation

        super().__init__(
            validations=[
                self.VALIDATION_EMAIL_PROTECTION_EKU_MISSING,
                self.VALIDATION_PROHIBITED_EKU_PRESENT,
            ],
            pdu_class=rfc5280.ExtKeyUsageSyntax,
        )

    def validate(self, node):
        kp_oids = set((kp.pdu for kp in node.children.values()))

        findings = []
        if rfc5280.id_kp_emailProtection not in kp_oids:
            findings.append(
                validation.ValidationFindingDescription(
                    self.VALIDATION_EMAIL_PROTECTION_EKU_MISSING, None
                )
            )

        if self._generation == smime_constants.Generation.STRICT:
            prohibited_kps = kp_oids.difference({rfc5280.id_kp_emailProtection})
        else:
            prohibited_kps = kp_oids.intersection(self._LEGACY_MP_PROHIBITED_EKUS)

        if len(prohibited_kps) > 0:
            findings.append(
                validation.ValidationFindingDescription(
                    self.VALIDATION_PROHIBITED_EKU_PRESENT,
                    f"Prohibited EKU(s) present: {oid.format_oids(prohibited_kps)}",
                )
            )

        return validation.ValidationResult(self, node, findings)


class AllowedKeyUsageValidator(validation.Validator):
    VALIDATION_UNKNOWN_CERT_TYPE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "cabf.smime.unknown_certificate_key_usage_type",
    )

    VALIDATION_REQUIRED_KU_MISSING = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR, "cabf.smime.required_ku_missing"
    )

    VALIDATION_PROHIBITED_KU_PRESENT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR, "cabf.smime.prohibited_ku_present"
    )

    VALIDATION_UNSUPPORTED_PUBKEY_TYPE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "cabf.smime.unsupported_public_key_type",
    )

    _ALL_KUS = {str(n) for n in rfc5280.KeyUsage.namedValues}

    def __init__(self, generation):
        super().__init__(
            validations=[
                self.VALIDATION_REQUIRED_KU_MISSING,
                self.VALIDATION_PROHIBITED_KU_PRESENT,
                self.VALIDATION_UNKNOWN_CERT_TYPE,
                self.VALIDATION_UNSUPPORTED_PUBKEY_TYPE,
            ],
            pdu_class=rfc5280.KeyUsage,
        )

        self._generation = generation

    @classmethod
    def _get_prohibited_kus(cls, node, allowed_kus):
        return {
            pk for pk in cls._ALL_KUS - allowed_kus if bitstring.has_named_bit(node, pk)
        }

    def validate(self, node):
        spki_alg_oid = node.navigate(
            ":certificate.tbsCertificate.subjectPublicKeyInfo.algorithm.algorithm"
        ).pdu

        allowed_kus = {KeyUsageBitName.DIGITAL_SIGNATURE}
        is_signing_cert = bitstring.has_named_bit(
            node, KeyUsageBitName.DIGITAL_SIGNATURE
        )

        if is_signing_cert:
            allowed_kus.add(KeyUsageBitName.NON_REPUDIATION)

        if spki_alg_oid == rfc3279.rsaEncryption:
            allowed_kus.add(KeyUsageBitName.KEY_ENCIPHERMENT)

            is_key_mgmt_cert = bitstring.has_named_bit(
                node, KeyUsageBitName.KEY_ENCIPHERMENT
            )

            if is_key_mgmt_cert and self._generation != Generation.STRICT:
                allowed_kus.add(KeyUsageBitName.DATA_ENCIPHERMENT)
        elif spki_alg_oid == rfc5480.id_ecPublicKey:
            is_key_mgmt_cert = bitstring.has_named_bit(
                node, KeyUsageBitName.KEY_AGREEMENT
            )

            if is_key_mgmt_cert:
                allowed_kus.update(
                    {
                        KeyUsageBitName.ENCIPHER_ONLY,
                        KeyUsageBitName.DECIPHER_ONLY,
                        KeyUsageBitName.KEY_AGREEMENT,
                    }
                )
        elif spki_alg_oid in {rfc8410.id_Ed448, rfc8410.id_Ed25519}:
            is_key_mgmt_cert = False
        else:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_UNSUPPORTED_PUBKEY_TYPE
            )

        if not is_signing_cert and not is_key_mgmt_cert:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_UNKNOWN_CERT_TYPE
            )

        prohibited_kus = self._get_prohibited_kus(node, allowed_kus)

        if len(prohibited_kus) > 0:
            ku_str = ", ".join(sorted(prohibited_kus))

            raise validation.ValidationFindingEncountered(
                self.VALIDATION_PROHIBITED_KU_PRESENT,
                f"Prohibited KUs present: {ku_str}",
            )


class EndEntityValidator(validation.Validator):
    VALIDATION_SMIME_CERT_IS_CA = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR, "cabf.smime.is_ca_certificate"
    )

    def __init__(self):
        super().__init__(
            validations=[self.VALIDATION_SMIME_CERT_IS_CA],
            pdu_class=rfc5280.BasicConstraints,
        )

    def validate(self, node):
        if bool(node.children["cA"].pdu):
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_SMIME_CERT_IS_CA
            )


class SubjectAlternativeNameContainsEmailAddressValidator(validation.Validator):
    VALIDATION_SAN_EMAIL_ADDRESS_MISSING = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "cabf.smime.san_does_not_contain_email_address",
    )

    def __init__(self):
        super().__init__(
            validations=[self.VALIDATION_SAN_EMAIL_ADDRESS_MISSING],
            pdu_class=rfc5280.SubjectAltName,
        )

    def validate(self, node):
        if len(get_email_addresses_from_san(node.document)) == 0:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_SAN_EMAIL_ADDRESS_MISSING
            )


class SubjectAlternativeNameProhibitedGeneralNameTypesValidator(validation.Validator):
    VALIDATION_PROHIBITED_GENERALNAME_TYPE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "cabf.smime.prohibited_generalname_type_present",
    )

    VALIDATION_PROHIBITED_OTHERNAME_TYPE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "cabf.smime.prohibited_othername_type_present",
    )

    _ALLOWED_TYPES = {"rfc822Name", "directoryName", "otherName"}

    def __init__(self, generation):
        super().__init__(
            validations=[
                self.VALIDATION_PROHIBITED_GENERALNAME_TYPE,
                self.VALIDATION_PROHIBITED_OTHERNAME_TYPE,
            ],
            pdu_class=rfc5280.SubjectAltName,
        )

        self._generation = generation

    def validate(self, node):
        present_types = set((gn.child[0] for gn in node.children.values()))

        prohibited_present_types = present_types - self._ALLOWED_TYPES

        if len(prohibited_present_types) > 0:
            type_str = ", ".join(prohibited_present_types)

            raise validation.ValidationFindingEncountered(
                self.VALIDATION_PROHIBITED_GENERALNAME_TYPE,
                f"Prohibited GeneralName type(s) present: {type_str}",
            )

        for gn in node.children.values():
            name, val = gn.child

            if name == "otherName":
                type_oid = val.children["type-id"].pdu

                if (
                    type_oid != rfc8398.id_on_SmtpUTF8Mailbox
                    and self._generation == Generation.STRICT
                ):
                    raise validation.ValidationFindingEncountered(
                        self.VALIDATION_PROHIBITED_OTHERNAME_TYPE,
                        f"Prohibited otherName present: {type_oid}",
                    )


class CabfSmimeKeyUsageCriticalityValidator(extension.ExtensionCriticalityValidator):
    VALIDATION_KU_NOT_CRITICAL = validation.ValidationFinding(
        validation.ValidationFindingSeverity.WARNING,
        "cabf.smime.ku_extension_not_critical",
    )

    def __init__(self):
        super().__init__(
            type_oid=rfc5280.id_ce_keyUsage,
            is_critical=True,
            validation=self.VALIDATION_KU_NOT_CRITICAL,
        )


class GmailAuthorityInfoAccessCaIssuersValidator(validation.Validator):
    VALIDATION_AIA_CA_ISSUERS_MISSING = validation.ValidationFinding(
        validation.ValidationFindingSeverity.NOTICE,
        "googl.gmail.authority_info_access_ca_issuers_missing",
    )

    def __init__(self):
        super().__init__(
            pdu_class=rfc5280.AuthorityInfoAccessSyntax,
            validations=[self.VALIDATION_AIA_CA_ISSUERS_MISSING],
        )

    def validate(self, node):
        if not any(
            (
                ad
                for ad in node.children.values()
                if ad.children["accessMethod"].pdu == rfc5280.id_ad_caIssuers
            )
        ):
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_AIA_CA_ISSUERS_MISSING
            )


class AllowedCrldpFullNameValidator(validation.Validator):
    VALIDATION_CRLDP_FULLNAME_PROHIBITED_GENERALNAME_TYPE = (
        validation.ValidationFinding(
            validation.ValidationFindingSeverity.ERROR,
            "cabf.smime.crldp_fullname_prohibited_generalname_type",
        )
    )

    VALIDATION_CRLDP_FULLNAME_PROHIBITED_URI_SCHEME = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "cabf.smime.crldp_fullname_prohibited_uri_scheme",
    )

    def __init__(self, generation):
        super().__init__(
            validations=[
                self.VALIDATION_CRLDP_FULLNAME_PROHIBITED_URI_SCHEME,
                self.VALIDATION_CRLDP_FULLNAME_PROHIBITED_GENERALNAME_TYPE,
            ],
            pdu_class=rfc5280.DistributionPointName,
            predicate=lambda n: "fullName" in n.children,
        )

        self._generation = generation

    def validate(self, node):
        uris = [
            str(gn.child[1].pdu)
            for gn in node.navigate("fullName").children.values()
            if "uniformResourceIdentifier" in gn.children
        ]

        if len(uris) != len(node.navigate("fullName").children):
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_CRLDP_FULLNAME_PROHIBITED_GENERALNAME_TYPE
            )

        allowed_schemes = (
            {"http", "ldap", "ftp"}
            if self._generation == Generation.LEGACY
            else {"http"}
        )
        for u in uris:
            scheme = urlparse(u).scheme

            if scheme.lower() not in allowed_schemes:
                raise validation.ValidationFindingEncountered(
                    self.VALIDATION_CRLDP_FULLNAME_PROHIBITED_URI_SCHEME,
                    f'Prohibited URI scheme: "{scheme}"',
                )


class AllowedAiaUriSchemeValidator(validation.Validator):
    VALIDATION_PROHIBITED_GENERALNAME_TYPE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "cabf.smime.aia_prohibited_generalname_type",
    )

    VALIDATION_PROHIBITED_URI_SCHEME = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "cabf.smime.aia_prohibited_uri_scheme",
    )

    def __init__(self, generation):
        super().__init__(
            validations=[
                self.VALIDATION_PROHIBITED_GENERALNAME_TYPE,
                self.VALIDATION_PROHIBITED_URI_SCHEME,
            ],
            pdu_class=rfc5280.AccessDescription,
            predicate=lambda n: n.parent is not None
            and isinstance(n.parent, rfc5280.AuthorityInfoAccessSyntax),
        )

        self._generation = generation

    def validate(self, node):
        desc_node = node.children["accessLocation"]
        name, value = desc_node.child

        if name != "uniformResourceIdentifier":
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_PROHIBITED_GENERALNAME_TYPE,
                f"Prohibited AIA GeneralName type: {name}",
            )

        scheme = urlparse(str(value.pdu)).scheme

        allowed_schemes = (
            {"http", "ftp", "ldap"}
            if self._generation == Generation.LEGACY
            else {"http"}
        )
        if scheme.lower() not in allowed_schemes:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_PROHIBITED_URI_SCHEME,
                f"Prohibited AIA URI scheme: {scheme}",
            )


class SubjectDirectoryAttributesPresenceValidator(validation.Validator):
    VALIDATION_SDA_EXTENSION_PROHIBITED = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "cabf.smime.subject_directory_attributes_extension_prohibited",
    )

    def __init__(self, validation_level, generation):
        super().__init__(
            validations=[self.VALIDATION_SDA_EXTENSION_PROHIBITED],
            pdu_class=rfc5280.SubjectDirectoryAttributes,
        )

        self._validation_level = validation_level
        self._generation = generation

    def validate(self, node):
        if (
            self._validation_level == ValidationLevel.MAILBOX
            or self._generation != Generation.LEGACY
        ):
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_SDA_EXTENSION_PROHIBITED
            )


class QCStatementsCriticalityValidator(extension.ExtensionCriticalityValidator):
    VALIDATION_QCS_EXTENSION_CRITICAL = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "cabf.smime.qc_statements_extension_critical",
    )

    def __init__(self):
        super().__init__(
            type_oid=rfc3739.id_pe_qcStatements,
            is_critical=False,
            validation=self.VALIDATION_QCS_EXTENSION_CRITICAL,
        )


class AdobeTimestampCriticalityValidator(extension.ExtensionCriticalityValidator):
    VALIDATION_ADOBE_TIMESTAMP_EXTENSION_CRITICAL = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "cabf.smime.adobe_timestamp_extension_critical",
    )

    def __init__(self):
        super().__init__(
            type_oid=adobe_asn1.id_adobe_timestamp,
            is_critical=False,
            validation=self.VALIDATION_ADOBE_TIMESTAMP_EXTENSION_CRITICAL,
        )


class AdobeArchiveRevInfoCriticalityValidator(extension.ExtensionCriticalityValidator):
    VALIDATION_ADOBE_ARCHIVE_REVINFO_EXTENSION_CRITICAL = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "cabf.smime.adobe_archive_revinfo_extension_critical",
    )

    def __init__(self):
        super().__init__(
            type_oid=adobe_asn1.id_adobe_archiverevinfo,
            is_critical=False,
            validation=self.VALIDATION_ADOBE_ARCHIVE_REVINFO_EXTENSION_CRITICAL,
        )


class AdobeTimestampPresenceValidator(validation.Validator):
    VALIDATION_ADOBE_TIMESTAMP_EXTENSION_PROHIBITED = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "cabf.smime.adobe_timestamp_extension_prohibited",
    )

    def __init__(self, generation):
        self._generation = generation

        super().__init__(
            validations=[self.VALIDATION_ADOBE_TIMESTAMP_EXTENSION_PROHIBITED],
            pdu_class=adobe_asn1.AdobeTimestamp,
        )

    def validate(self, node):
        if self._generation == smime_constants.Generation.STRICT:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_ADOBE_TIMESTAMP_EXTENSION_PROHIBITED
            )


class AdobeArchiveRevInfoPresenceValidator(validation.Validator):
    VALIDATION_ADOBE_ARCHIVE_REVINFO_EXTENSION_PROHIBITED = (
        validation.ValidationFinding(
            validation.ValidationFindingSeverity.ERROR,
            "cabf.smime.adobe_archive_revinfo_extension_prohibited",
        )
    )

    def __init__(self, generation):
        self._generation = generation

        super().__init__(
            validations=[self.VALIDATION_ADOBE_ARCHIVE_REVINFO_EXTENSION_PROHIBITED],
            pdu_class=adobe_asn1.AdobeArchiveRevInfo,
        )

    def validate(self, node):
        if self._generation == smime_constants.Generation.STRICT:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_ADOBE_ARCHIVE_REVINFO_EXTENSION_PROHIBITED
            )


class CrlDistributionPointPresenceValidator(extension.ExtensionPresenceValidator):
    VALIDATION_CRLDP_MISSING = validation.ValidationFinding(
        validation.ValidationFindingSeverity.WARNING,
        "cabf.smime.crldp_extension_missing",
    )

    def __init__(self):
        super().__init__(
            extension_oid=rfc5280.id_ce_cRLDistributionPoints,
            validation=self.VALIDATION_CRLDP_MISSING,
            pdu_class=rfc5280.Extensions,
        )



import functools
import logging
from typing import Set, Optional

from cryptography import x509, exceptions
from cryptography.hazmat.primitives.asymmetric import (
    padding,
    rsa,
    dsa,
    ec,
    ed25519,
    ed448,
)
from pyasn1.codec.der.encoder import encode
from pyasn1.type import univ
from pyasn1.type.base import Asn1Type
from pyasn1_alt_modules import rfc5280, rfc3739

from pkilint import validation, pkix, document
from pkilint.document import Document, ValueDecoder
from pkilint.pkix import (
    extension,
    time,
    name,
    create_name_validator_container,
    general_name,
    algorithm,
)
from pkilint.pkix.certificate import (
    certificate_validity,
    certificate_extension,
    certificate_validator,
    certificate_key,
    certificate_name,
    certificate_transparency,
)

logger = logging.getLogger(__name__)


class RFC5280Certificate(Document):
    def __init__(self, substrate_source, substrate, name=None, parent=None):
        super().__init__(
            rfc5280.Certificate(), substrate_source, substrate, name, parent
        )

    @property
    def not_before(self):
        try:
            return time.parse_time_node(
                self.root.navigate("tbsCertificate.validity.notBefore")
            )
        except ValueError:
            return pkix.MAXIMUM_TIME_DATETIME

    @property
    def not_after(self):
        try:
            return time.parse_time_node(
                self.root.navigate("tbsCertificate.validity.notAfter")
            )
        except ValueError:
            return pkix.MAXIMUM_TIME_DATETIME

    def _decode_and_append_extension(
        self, ext_oid: univ.ObjectIdentifier, ext_asn1_spec: Asn1Type
    ) -> Optional[document.PDUNode]:
        ext_and_idx = self.get_extension_by_oid(ext_oid)

        if ext_and_idx is None:
            return None

        ext, _ = ext_and_idx
        ext_value = ext.children["extnValue"]

        try:
            return document.decode_substrate(
                self, ext_value.pdu.asOctets(), ext_asn1_spec, ext_value
            )
        except ValueError:
            # suppress decoding errors, which will be reported by DecodingValidator instances
            return None

    @functools.cached_property
    def is_ca(self) -> bool:
        decoded = self._decode_and_append_extension(
            rfc5280.id_ce_basicConstraints, rfc5280.BasicConstraints()
        )

        return bool(decoded.navigate("cA").pdu) if decoded else False

    @functools.cached_property
    def extended_key_usages(self) -> Set[univ.ObjectIdentifier]:
        decoded = self._decode_and_append_extension(
            rfc5280.id_ce_extKeyUsage, rfc5280.ExtKeyUsageSyntax()
        )

        return (
            {eku_node.pdu for eku_node in decoded.children.values()}
            if decoded
            else set()
        )

    @functools.cached_property
    def qualified_statement_ids(self) -> Set[univ.ObjectIdentifier]:
        decoded = self._decode_and_append_extension(
            rfc3739.id_pe_qcStatements, rfc3739.QCStatements()
        )

        return (
            {qs.children["statementId"].pdu for qs in decoded.children.values()}
            if decoded
            else set()
        )

    @functools.cached_property
    def cryptography_object(self):
        return x509.load_der_x509_certificate(self.substrate)

    @functools.cached_property
    def is_self_issued(self):
        issuer_node = self.root.navigate("tbsCertificate.issuer")
        subject_node = self.root.navigate("tbsCertificate.subject")

        return encode(issuer_node.pdu) == encode(subject_node.pdu)

    @functools.cached_property
    def is_self_signed(self):
        if not self.is_self_issued:
            return False

        public_key = self.cryptography_object.public_key()
        tbs_octets = self.cryptography_object.tbs_certificate_bytes
        hash_alg = self.cryptography_object.signature_hash_algorithm
        signature_octets = self.cryptography_object.signature

        try:
            if isinstance(public_key, rsa.RSAPublicKey):
                public_key.verify(
                    signature_octets, tbs_octets, padding.PKCS1v15(), hash_alg
                )
            elif isinstance(public_key, dsa.DSAPublicKey):
                public_key.verify(signature_octets, tbs_octets, hash_alg)
            elif isinstance(public_key, ec.EllipticCurvePublicKey):
                public_key.verify(signature_octets, tbs_octets, ec.ECDSA(hash_alg))
            elif isinstance(public_key, ed25519.Ed25519PublicKey):
                public_key.verify(signature_octets, tbs_octets)
            elif isinstance(public_key, ed448.Ed448PublicKey):
                public_key.verify(signature_octets, tbs_octets)
        except exceptions.InvalidSignature:
            return False

        return True

    def get_extension_by_oid(self, oid):
        tbs_cert = self.root.children["tbsCertificate"]

        # ensure there's extensions
        if "extensions" not in tbs_cert.children:
            return None

        extensions_node = tbs_cert.children["extensions"]

        for ext_idx, extension_node in extensions_node.children.items():
            ext_oid = extension_node.children["extnID"].pdu

            if ext_oid == oid:
                return extension_node, int(ext_idx)

        return None

    def get_name_attributes_by_type(self, oid, name_path):
        name_node = self.root.navigate(name_path)

        return name.get_name_attributes_by_type(name_node, oid)

    def get_issuer_attributes_by_type(self, oid):
        return self.get_name_attributes_by_type(oid, "tbsCertificate.issuer")

    def get_subject_attributes_by_type(self, oid):
        return self.get_name_attributes_by_type(oid, "tbsCertificate.subject")

    @functools.cached_property
    def policy_oids(self) -> Set[univ.ObjectIdentifier]:
        decoded = self._decode_and_append_extension(
            rfc5280.id_ce_certificatePolicies, rfc5280.CertificatePolicies()
        )

        return (
            {pi.children["policyIdentifier"].pdu for pi in decoded.children.values()}
            if decoded
            else set()
        )


def create_spki_decoder(
    subject_public_key_type_mappings, subject_public_key_parameters_type_mappings
):
    subject_public_key_decoder = certificate_key.SubjectPublicKeyDecoder(
        type_mappings=subject_public_key_type_mappings
    )
    subject_public_key_parameters_decoder = (
        certificate_key.SubjectPublicKeyParametersDecoder(
            type_mappings=subject_public_key_parameters_type_mappings
        )
    )

    return validation.ValidatorContainer(
        validators=[
            certificate_key.SubjectPublicKeyDecodingValidator(
                decode_func=subject_public_key_decoder
            ),
            certificate_key.SubjectPublicKeyParametersDecodingValidator(
                decode_func=subject_public_key_parameters_decoder
            ),
        ],
        pdu_class=rfc5280.SubjectPublicKeyInfo,
    )


def create_policy_qualifier_decoder(type_mappings):
    decoder = ValueDecoder(
        type_path="policyQualifierId",
        value_path="qualifier",
        type_mappings=type_mappings,
    )

    return validation.DecodingValidator(
        decode_func=decoder, pdu_class=rfc5280.PolicyQualifierInfo
    )


def create_other_name_decoder(type_mappings):
    decoder = ValueDecoder(
        type_path="type-id", value_path="value", type_mappings=type_mappings
    )

    return validation.DecodingValidator(
        decode_func=decoder, pdu_class=rfc5280.AnotherName
    )


def create_qc_statements_decoder(type_mappings):
    decoder = ValueDecoder(
        type_path="statementId", value_path="statementInfo", type_mappings=type_mappings
    )

    return validation.DecodingValidator(
        decode_func=decoder, pdu_class=rfc3739.QCStatement
    )


def create_issuer_validator_container(additional_validators=None, **kwargs):
    if additional_validators is None:
        additional_validators = []

    if len(kwargs) == 0:
        kwargs["path"] = "certificate.tbsCertificate.issuer"

    return create_name_validator_container(
        [
            name.EmptyNameValidator(),
            general_name.MailboxAddressSyntaxValidator(pdu_class=rfc5280.EmailAddress),
        ]
        + additional_validators,
        **kwargs
    )


def create_subject_validator_container(additional_validators=None, **kwargs):
    if additional_validators is None:
        additional_validators = []

    if len(kwargs) == 0:
        kwargs["path"] = "certificate.tbsCertificate.subject"

    return create_name_validator_container(
        [
            certificate_name.SubjectEmailAddressInSanValidator(),
            general_name.MailboxAddressSyntaxValidator(pdu_class=rfc5280.EmailAddress),
            name.DomainComponentValidDomainNameValidator(pdu_class=rfc5280.Name),
        ]
        + additional_validators,
        **kwargs
    )


def create_validity_validator_container(additional_validators=None):
    if additional_validators is None:
        additional_validators = []
    return validation.ValidatorContainer(
        validators=[
            certificate_validity.CertificateSaneValidityPeriodValidator(),
            time.TimeCorrectEncodingValidator(),
        ]
        + additional_validators,
        path="certificate.tbsCertificate.validity",
    )


def create_extensions_validator_container(additional_validators=None):
    if additional_validators is None:
        additional_validators = []
    return validation.ValidatorContainer(
        validators=[
            extension.UniqueExtensionValidator(),
            certificate_extension.BasicConstraintsValidator(),
            certificate_extension.CertificatePolicyQualifierValidator(),
            certificate_extension.CertificatePoliciesUserNoticeValidator(),
            certificate_key.SubjectKeyIdentifierValidator(),
            certificate_extension.SubjectKeyIdentifierCriticalityValidator(),
            certificate_extension.KeyUsageCriticalityValidator(),
            certificate_extension.KeyUsageValidator(),
            general_name.UriSyntaxValidator(pdu_class=rfc5280.CPSuri),
            general_name.GeneralNameValidatorContainer(),
            certificate_extension.DuplicatePolicyValidator(),
            certificate_extension.CrlDpCriticalityValidator(),
            certificate_extension.NameConstraintsCriticalityValidator(),
            extension.AuthorityKeyIdentifierValidator(),
            extension.AuthorityKeyIdentifierCriticalityValidator(),
            certificate_extension.NameConstraintsValidator(),
            certificate_extension.NameConstraintsGeneralSubtreeValidator(),
            certificate_extension.AuthorityInformationAccessCriticalityValidator(),
            certificate_extension.SubjectInformationAccessCriticalityValidator(),
            certificate_extension.SubjectAlternativeNameCriticalityValidator(),
            certificate_extension.SubjectDirectoryAttributesCriticalityValidator(),
            certificate_extension.SmimeCapabilitiesCriticalityValidator(),
            extension.DistributionPointValidator(),
            certificate_extension.CtPrecertPoisonCriticalityValidator(),
            certificate_extension.CtPrecertPoisonSctListMutuallyExclusiveExtensionsValidator(),
            certificate_transparency.SctListElementCountValidator(),
            certificate_extension.PolicyConstraintsPresenceValidator(),
            certificate_extension.PolicyMappingsPresenceValidator(),
            certificate_extension.InhibitAnyPolicyPresenceValidator(),
            certificate_extension.ProhibitedQualifiedStatementValidator(),
            certificate_extension.IssuerAlternativeNameCriticalityValidator(),
        ]
        + additional_validators,
        path="certificate.tbsCertificate.extensions",
    )


def create_pkix_certificate_validator_container(decoding_validators, validators):
    decoding_validator_container = [
        validation.ValidatorContainer(
            validators=decoding_validators, path="certificate"
        )
    ]

    validators += [
        certificate_validator.CorrectVersionValidator(),
        pkix.CertificateSerialNumberValidator(),
        certificate_validator.SignatureAlgorithmMatchValidator(),
        certificate_extension.SubjectKeyIdentifierPresenceValidator(),
        certificate_extension.AuthorityKeyIdentifierPresenceValidator(),
        certificate_extension.KeyUsagePresenceValidator(),
        time.UtcTimeCorrectSyntaxValidator(),
        time.GeneralizedTimeCorrectSyntaxValidator(),
        certificate_validator.IssuerUniqueIdAbsenceValidator(),
        certificate_validator.SubjectUniqueIdAbsenceValidator(),
    ]

    return validation.ValidatorContainer(
        validators=decoding_validator_container + validators
    )


def create_decoding_validators(
    name_mappings, extension_mappings, additional_decoding_validators=None
):
    if additional_decoding_validators is None:
        additional_decoding_validators = []
    return [
        pkix.create_attribute_decoder(name_mappings),
        pkix.create_extension_decoder(extension_mappings),
        pkix.create_signature_algorithm_identifier_decoder(
            algorithm.SIGNATURE_ALGORITHM_IDENTIFIER_MAPPINGS,
            path="certificate.tbsCertificate.signature",
        ),
        create_spki_decoder(
            certificate_key.SUBJECT_PUBLIC_KEY_ALGORITHM_IDENTIFIER_MAPPINGS,
            certificate_key.SUBJECT_KEY_PARAMETER_ALGORITHM_IDENTIFIER_MAPPINGS,
        ),
        create_policy_qualifier_decoder(
            certificate_extension.CERTIFICATE_POLICY_QUALIFIER_MAPPINGS
        ),
        create_other_name_decoder(general_name.OTHER_NAME_MAPPINGS),
        certificate_transparency.SctListExtensionDecodingValidator(),
    ] + additional_decoding_validators



import re

from iso3166 import countries_by_alpha2
from pyasn1_alt_modules import rfc3739, rfc5280

import pkilint.oid
from pkilint import validation
from pkilint.etsi import etsi_constants
from pkilint.etsi.asn1 import ts_119_495 as ts_119_495_asn1
from pkilint.itu import x520_name


class RolesOfPspValidator(validation.Validator):
    """
    GEN-5.2.2-1: RolesOfPSP shall contain one or more roles or contain a single entry indicating that the role is
    unspecified.
    GEN-5.2.2-2 If the certificate is issued for EU PSD2 the role object identifier shall be the
    appropriate one of the four OIDS defined in the ASN.1 snippet below; and
    GEN-5.2.2-3 If the certificate is issued for EU PSD2 the role name shall be  the appropriate one of the abbreviated
    names defined in clause 5.1 PSP_AS, PSP_PI, or PSP_IC.
    GEN-5.2.2-3A If the role is unspecified the role name shall be "Unspecified"
    GEN-5.2.2-5 The TSP shall ensure that the name in roleofPSPName is the one associated with the role object
    identifier held in roleofPSPOid.
    """

    VALIDATION_PSP_ROLES_EMPTY = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "etsi.ts_119_495.gen-5.2.2-1.roles_of_psp_empty",
    )
    VALIDATION_PSP_ROLES_INVALID = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "etsi.ts_119_495.gen-5.2-2.invalid_psp_role",
    )

    VALIDATION_PSP_OIDS_INVALID = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "etsi.ts_119_495.gen-5.2-2.invalid_psp_oid",
    )

    VALIDATION_PSP_ROLES_MISMATCH = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "etsi.ts_119_495.gen-5.2.2-5.psp_role_mismatch",
    )

    def __init__(self):
        super().__init__(
            validations=[
                self.VALIDATION_PSP_ROLES_EMPTY,
                self.VALIDATION_PSP_ROLES_INVALID,
                self.VALIDATION_PSP_ROLES_MISMATCH,
                self.VALIDATION_PSP_OIDS_INVALID,
            ],
            pdu_class=ts_119_495_asn1.RolesOfPSP,
        )
        self._expected_roles = {
            ts_119_495_asn1.id_psd2_role_psp_ai: "PSP_AI",
            ts_119_495_asn1.id_psd2_role_psp_as: "PSP_AS",
            ts_119_495_asn1.id_psd2_role_psp_ic: "PSP_IC",
            ts_119_495_asn1.id_psd2_role_psp_pi: "PSP_PI",
            ts_119_495_asn1.id_psd2_role_psp_unspecified: "Unspecified",
        }

    def validate(self, node):
        if not any(node.children):
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_PSP_ROLES_EMPTY
            )

        for child in node.children.values():
            psp_oid = child.children["roleOfPspOid"].pdu
            role_psp = str(child.children["roleOfPspName"].pdu)
            expected_role = self._expected_roles.get(psp_oid)

            if psp_oid not in self._expected_roles.keys():
                raise validation.ValidationFindingEncountered(
                    self.VALIDATION_PSP_OIDS_INVALID,
                    f"expected oid values are {pkilint.oid.format_oids(self._expected_roles.keys())} got {psp_oid}",
                )
            if role_psp not in self._expected_roles.values():
                raise validation.ValidationFindingEncountered(
                    self.VALIDATION_PSP_ROLES_INVALID,
                    f"expected role values are [ {', '.join(map(str, self._expected_roles.values()))}]. Got {role_psp}",
                )
            if role_psp != expected_role:
                raise validation.ValidationFindingEncountered(
                    self.VALIDATION_PSP_ROLES_MISMATCH,
                    f"Expected role is: {expected_role}. Role in cert is: {role_psp}. Oid in cert is: {psp_oid}",
                )


class PresenceofQCEUPDSStatementValidator(validation.Validator):
    """GEN-5.1.1 The Open Banking Attributes shall be included in a QCSTatement within the qcStatements extension
    as specified in clause 3.2.5 of IETF RFC 3739."""

    VALIDATION_QC_EU_PDS_MISSING = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "etsi.ts_119_495.gen-5.1.1.qc_eu_pds_missing",
    )

    def __init__(self):
        super().__init__(
            validations=[self.VALIDATION_QC_EU_PDS_MISSING],
            pdu_class=rfc3739.QCStatements,
        )

    def validate(self, node):
        if (
            ts_119_495_asn1.id_etsi_psd2_qcStatement
            not in node.document.qualified_statement_ids
        ):
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_QC_EU_PDS_MISSING
            )


class NCANameLatinCharactersValidator(validation.Validator):
    """GEN-5.2.3-1: The NCAName shall be plain text using Latin alphabet provided by the Competent Authority itself
    for purpose of identification in certificates."""

    VALIDATION_NCA_NAME_NON_LATIN = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "etsi.ts_119_495.gen-5.2.3-1.nca_name_non_latin",
    )

    def __init__(self):
        super().__init__(
            validations=[self.VALIDATION_NCA_NAME_NON_LATIN],
            pdu_class=ts_119_495_asn1.NCAName,
        )

    def validate(self, node):
        nca_name = str(node.pdu)

        if not nca_name.isascii():
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_NCA_NAME_NON_LATIN, f"invalid NCA name: {nca_name}"
            )


class NCAIdValidator(validation.Validator):
    """GEN-5.2.3-2: Validator for NCAId structure.
    The NCAId shall contain information using the following structure in the presented order:
    • 2 character ISO 3166-1 country code representing the Competent Authority country;
    • hyphen-minus "-" (0x2D (ASCII), U+002D (UTF-8)); and
    • 2-8 character Competent Authority identifier without country code (A-Z uppercase only, no separator).
    """

    VALIDATION_INVALID_STRUCTURE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "etsi.ts_119_495.gen-5.2.3-2.invalid_structure",
    )
    VALIDATION_INVALID_ISO_COUNTRY = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "etsi.ts_119_495.gen-5.2.3-2.invalid_iso_country",
    )
    VALIDATION_INVALID_CA_IDENTIFIER = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "etsi.ts_119_495.gen-5.2.3-2.invalid_ca_identifier",
    )

    def __init__(self):
        super().__init__(
            validations=[
                self.VALIDATION_INVALID_STRUCTURE,
                self.VALIDATION_INVALID_ISO_COUNTRY,
                self.VALIDATION_INVALID_CA_IDENTIFIER,
            ],
            pdu_class=ts_119_495_asn1.NCAId,
        )

    def validate(self, node):
        nca_id = str(node.pdu)

        if nca_id.count("-") != 1:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_INVALID_STRUCTURE,
                f"Invalid separator in NCAId: {nca_id}",
            )

        iso_country_code, ca_identifier = nca_id.rsplit("-", 1)

        if iso_country_code not in countries_by_alpha2:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_INVALID_ISO_COUNTRY,
                f"Invalid ISO country code: {iso_country_code}",
            )

        if not (
            2 <= len(ca_identifier) <= 8
            and ca_identifier.isalpha()
            and ca_identifier.isupper()
        ):
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_INVALID_CA_IDENTIFIER,
                f"Invalid Competent Authority identifier: {ca_identifier}",
            )


class PsdOrganizationIdentifierFormatValidator(validation.Validator):
    """
    GEN-5.2.1-3: If an Authorization Number was issued by a Competent Authority the subject organizationIdentifier
    attribute should contain the Authorization Number encoded using the following structure in the presented order:

    • "PSD" as 3 character legal person identity type reference;
    • 2 character ISO 3166-1 [8] country code representing the Competent Authority country;
    • hyphen-minus "-" (0x2D (ASCII), U+002D (UTF-8));
    • 2-8 character Competent Authority identifier without country code (A-Z uppercase only, no separator);
    • hyphen-minus "-" (0x2D (ASCII), U+002D (UTF-8)); and
    • identifier (authorization number as specified by the Competent Authority. There are no restrictions on the
    characters used).

    ...

    GEN-5.3-3: The organizationIdentifier shall be present in the Subject's Distinguished Name and encoded with legal
    person syntax as specified in clause 5.2.1.
    """

    VALIDATION_INVALID_PSD_ORGANIZATION_ID_FORMAT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "etsi.ts_119_495.gen-5.2.1-3.invalid_psd_organization_id_format",
    )

    _PSD_ORGID_FORMAT_REGEX = re.compile("^PSD[A-Z]{2}-[A-Z]{2,8}-.+$")

    def __init__(self):
        super().__init__(
            validations=[self.VALIDATION_INVALID_PSD_ORGANIZATION_ID_FORMAT],
            pdu_class=x520_name.X520OrganizationIdentifier,
        )

    def validate(self, node):
        try:
            _, decoded_value_node = node.child
        except ValueError:
            return

        value_str = str(decoded_value_node.pdu)

        m = self._PSD_ORGID_FORMAT_REGEX.match(value_str)

        if m is None:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_INVALID_PSD_ORGANIZATION_ID_FORMAT,
                f'Invalid PSD organization identifier format: "{value_str}"',
            )


class Psd2CertificatePolicyOidPresenceValidator(validation.Validator):
    """
    OVR-6.1-3: TSPs issuing certificates for EU PSD2 may use the following policy identifier to augment the policy
    requirements associated with policy identifier QEVCP-w or QNCP-w as specified in ETSI EN 319 411-2 [5] giving
    precedence to the requirements defined in the present document.
    """

    VALIDATION_PROHIBITED_PSD2_POLICY_OID_PRESENT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "etsi.ts_119_495.ovr-6.1-3.prohibited_psd2_policy_oid_present",
    )

    def __init__(self, certificate_type):
        super().__init__(
            validations=[self.VALIDATION_PROHIBITED_PSD2_POLICY_OID_PRESENT],
            pdu_class=rfc5280.CertificatePolicies,
        )

        self._certificate_type = certificate_type

    def validate(self, node):
        if (
            ts_119_495_asn1.qcp_web_psd2 in node.document.policy_oids
            and self._certificate_type
            not in etsi_constants.PSD2_EIDAS_CERTIFICATE_TYPES
        ):
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_PROHIBITED_PSD2_POLICY_OID_PRESENT,
                f'Certificate type is "{self._certificate_type}" but PSD2 policy identifier is present',
            )
